# yellowbrick.features.projection
# Base class for all projection (decomposition) high dimensional data visualizers.
#
# Author:   Naresh Bachwani
# Created:  Wed Jul 17 08:59:33 2019 -0400
#
# Copyright (C) 2019, the scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: projection.py [21eb9d2] 43993586+naresh-bachwani@users.noreply.github.com $

"""
Base class for all projection (decomposition) high dimensional data visualizers.
"""

##########################################################################
## Imports
##########################################################################

import warnings
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import mpl_toolkits.mplot3d  # noqa

from yellowbrick.draw import manual_legend
from yellowbrick.features.base import DataVisualizer, TargetType
from yellowbrick.exceptions import YellowbrickValueError, YellowbrickWarning, NotFitted


##########################################################################
## Projection Visualizers
##########################################################################


class ProjectionVisualizer(DataVisualizer):
    """
    The ProjectionVisualizer provides functionality for projecting a multi-dimensional
    dataset into either 2 or 3 components so they can be plotted as a scatter plot on
    2d or 3d axes. The visualizer acts as a transformer, and draws the transformed data
    on behalf of the user. Because it is a DataVisualizer, the ProjectionVisualizer
    can plot continuous scatter plots with a colormap or discrete scatter plots with
    a legend.

    This visualizer is a base class and is not intended to be uses directly.
    Subclasses should implement a ``transform()`` method that calls ``draw()`` using
    the transformed data and the optional target as input.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes.
        will be used (or generated if required).

    features : list, default: None
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    classes : list, default: None
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised. This parameter is only used in the
        discrete target type case and is ignored otherwise.

    colors : list or tuple, default: None
        A single color to plot all instances as or a list of colors to color each
        instance according to its class in the discrete case or as an ordered
        colormap in the sequential case. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        The colormap used to create the individual colors. In the discrete case
        it is used to compute the number of colors needed for each class and
        in the continuous case it is used to create a sequential color map based
        on the range of the target.

    target_type : str, default: "auto"
        Specify the type of target as either "discrete" (classes) or "continuous"
        (real numbers, usually for regression). If "auto", then it will
        attempt to determine the type by counting the number of unique values.

        If the target is discrete, the colors are returned as a dict with classes
        being the keys. If continuous the colors will be list having value of
        color for each point. In either case, if no target is specified, then
        color will be specified as the first color in the color cycle.

    projection : int or string, default: 2
        The number of axes to project into, either 2d or 3d. To plot 3d plots
        with matplotlib, please ensure a 3d axes is passed to the visualizer,
        otherwise one will be created using the current figure.

    alpha : float, default: 0.75
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    colorbar : bool, default: True
        If the target_type is "continous" draw a colorbar to the right of the
        scatter plot. The colobar axes is accessible using the cax property.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.
    """

    def __init__(
        self,
        ax=None,
        features=None,
        classes=None,
        colors=None,
        colormap=None,
        target_type="auto",
        projection=2,
        alpha=0.75,
        colorbar=True,
        **kwargs
    ):

        super(ProjectionVisualizer, self).__init__(
            ax=ax,
            features=features,
            classes=classes,
            colors=colors,
            colormap=colormap,
            target_type=target_type,
            **kwargs
        )

        # Convert string to integer
        if isinstance(projection, str):
            if projection in {"2D", "2d"}:
                projection = 2
            if projection in {"3D", "3d"}:
                projection = 3
        if projection not in {2, 3}:
            raise YellowbrickValueError("Projection dimensions must be either 2 or 3")
        self.projection = projection

        if self.ax.name != "3d" and self.projection == 3:
            warnings.warn(
                "data projection to 3 dimensions requires a 3d axes to draw on.",
                YellowbrickWarning,
            )

        self.alpha = alpha
        self.colorbar = colorbar
        self._cax = None

    @property
    def cax(self):
        """
        The axes of the colorbar, right of the scatterplot.
        """
        if self._cax is None:
            raise AttributeError("This visualizer does not have an axes for colorbar")

        return self._cax

    @property
    def ax(self):
        """
        Overloads the axes property from base class. If no axes is specified then
        creates an axes for users. A 3d axes is created for 3 dimensional plots.
        """
        if not hasattr(self, "_ax") or self._ax is None:
            if self.projection == 3:
                fig = plt.gcf()
                self._ax = fig.add_subplot(111, projection="3d")
            else:
                self._ax = plt.gca()
        return self._ax

    @ax.setter
    def ax(self, ax):
        self._ax = ax

    def layout(self, divider=None):
        """
        Creates the layout for colorbar when target type is continuous.
        The colorbar is added to the right of the scatterplot.

        Subclasses can override this method to add other axes or layouts.

        Parameters
        ----------
        divider: AxesDivider
            An AxesDivider to be passed among all layout calls.
        """
        if (
            self._target_color_type == TargetType.CONTINUOUS
            and self.projection == 2
            and self.colorbar
            and self._cax is None
        ):
            # Ensure matplotlib version compatibility
            if make_axes_locatable is None:
                raise YellowbrickValueError(
                    (
                        "Colorbar requires matplotlib 2.0.2 or greater "
                        "please upgrade matplotlib"
                    )
                )

            # Create the new axes for the colorbar
            if divider is None:
                divider = make_axes_locatable(self.ax)

            self._cax = divider.append_axes("right", size="5%", pad=0.3)
            self._cax.set_yticks([])
            self._cax.set_xticks([])

    def fit_transform(self, X, y=None):
        """
        Fits the visualizer on the input data, and returns transformed X.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix or data frame of n instances with m features where m>2.

        y : array-like of shape (n,), optional
            A vector or series with target values for each instance in X. This
            vector is used to determine the color of the points in X.

        Returns
        -------
        Xprime : array-like of shape (n, 2)
            Returns the 2-dimensional embedding of the instances.
        """
        return self.fit(X, y).transform(X, y)

    def draw(self, Xp, y=None):
        """
        Draws the points described by Xp and colored by the points in y. Can be
        called multiple times before finalize to add more scatter plots to the
        axes, however ``fit()`` must be called before use.

        Parameters
        ----------
        Xp : array-like of shape (n, 2) or (n, 3)
            The matrix produced by the ``transform()`` method.

        y : array-like of shape (n,), optional
            The target, used to specify the colors of the points.

        Returns
        -------
        self.ax : matplotlib Axes object
            Returns the axes that the scatter plot was drawn on.
        """
        scatter_kwargs = self._determine_scatter_kwargs(y)

        # Draws the layout of the visualizer. It draws the axes for colorbars,
        # heatmap, etc.
        self.layout()

        if self.projection == 2:
            # Adds colorbar axis for continuous target type.
            self.ax.scatter(Xp[:, 0], Xp[:, 1], **scatter_kwargs)

        if self.projection == 3:
            self.ax.scatter(Xp[:, 0], Xp[:, 1], Xp[:, 2], **scatter_kwargs)

        return self.ax

    def finalize(self):
        """
        Draws legends and colorbar for scatter plots.
        """
        self.ax.set_xticklabels([])
        self.ax.set_yticklabels([])
        if self.projection == 3:
            self.ax.set_zticklabels([])

        if self._target_color_type == TargetType.DISCRETE:
            # Add the legend
            manual_legend(
                self, self.classes_, list(self._colors.values()), frameon=True
            )
        elif self._target_color_type == TargetType.CONTINUOUS:
            if self.colorbar:
                if self.projection == 3:
                    sm = plt.cm.ScalarMappable(cmap=self._colors, norm=self._norm)
                    # Avoid MPL TypeError: "You must first set_array for mappable"
                    sm.set_array([])
                    self.cbar = plt.colorbar(sm, ax=self.ax)

                else:
                    # Manually draw the colorbar.
                    self.cbar = mpl.colorbar.ColorbarBase(
                        self.cax, cmap=self._colors, norm=self._norm
                    )

    def _determine_scatter_kwargs(self, y=None):
        """
        Determines scatter argumnets to pass into ``plt.scatter()``. If y is
        discrete or single then determine colors. If continuous then determine
        colors and colormap.Also normalize to range

        Parameters
        ----------
        y : array-like of shape (n,), optional
            The target, used to specify the colors of the points for continuous
            target.
        """

        scatter_kwargs = {"alpha": self.alpha}
        # Determine the colors
        if self._target_color_type == TargetType.SINGLE:
            scatter_kwargs["c"] = self._colors

        elif self._target_color_type == TargetType.DISCRETE:
            if y is None:
                raise YellowbrickValueError("y is required for discrete target")

            try:
                scatter_kwargs["c"] = [self._colors[self.classes_[yi]] for yi in y]
            except IndexError:
                raise YellowbrickValueError("Target needs to be label encoded.")

        elif self._target_color_type == TargetType.CONTINUOUS:
            if y is None:
                raise YellowbrickValueError("y is required for continuous target")

            scatter_kwargs["c"] = y
            scatter_kwargs["cmap"] = self._colors
            self._norm = mpl.colors.Normalize(vmin=self.range_[0], vmax=self.range_[1])

        else:
            # Technically this should never be raised
            raise NotFitted("could not determine target color type")
        return scatter_kwargs



# yellowbrick.regressor.influence
# Visualize the influence and leverage of individual instances on a regression model.
#
# Author:   Benjamin Bengfort
# Created:  Sun Jun 09 15:21:17 2019 -0400
#
# Copyright (C) 2019 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: influence.py [fe14cfd] benjamin@bengfort.com $

"""
Visualize the influence and leverage of individual instances on a regression model.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import scipy as sp

from yellowbrick.base import Visualizer
from sklearn.linear_model import LinearRegression


##########################################################################
## Cook's Distance
##########################################################################


class CooksDistance(Visualizer):
    """
    Cook's Distance is a measure of how influential an instance is to the computation of
    a regression, e.g. if the instance is removed would the estimated coeficients of the
    underlying model be substantially changed? Because of this, Cook's Distance is
    generally used to detect outliers in standard, OLS regression. In fact, a general
    rule of thumb is that D(i) > 4/n is a good threshold for determining highly
    influential points as outliers and this visualizer can report the percentage of data
    that is above that threshold.

    This implementation of Cook's Distance assumes Ordinary Least Squares regression,
    and therefore embeds a ``sklearn.linear_model.LinearRegression`` under the hood.
    Distance is computed via the non-whitened leverage of the projection matrix,
    computed inside of ``fit()``. The results of this visualizer are therefore similar
    to, but not as advanced, as a similar computation using statsmodels. Computing the
    influence for other regression models requires leave one out validation and can be
    expensive to compute.

    .. seealso::
        For a longer discussion on detecting outliers in regression and computing
        leverage and influence, see `linear regression in python, outliers/leverage
        detect <http://bit.ly/2If2fga>`_ by Huiming Song.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    draw_threshold : bool, default: True
        Draw a horizontal line at D(i) == 4/n to easily identify the most influential
        points on the final regression. This will also draw a legend that specifies the
        percentage of data points that are above the threshold.

    linefmt : str, default: 'C0-'
        A string defining the properties of the vertical lines of the stem plot, usually
        this will be a color or a color and a line style. The default is simply a solid
        line with the first color of the color cycle.

    markerfmt : str, default: ','
        A string defining the properties of the markers at the stem plot heads. The
        default is "pixel", e.g. basically no marker head at the top of the stem plot.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence the final
        visualization (e.g. size or title parameters).

    Attributes
    ----------
    distance_ : array, 1D
        The Cook's distance value for each instance specified in ``X``, e.g. an 1D array
        with shape ``(X.shape[0],)``.

    p_values_ : array, 1D
        The p values associated with the F-test of Cook's distance distribution. A 1D
        array whose shape matches ``distance_``.

    influence_threshold_ : float
        A rule of thumb influence threshold to determine outliers in the regression
        model, defined as It=4/n.

    outlier_percentage_ : float
        The percentage of instances whose Cook's distance is greater than the influnce
        threshold, the percentage is 0.0 <= p  <= 100.0.

    Notes
    -----
    Cook's Distance is very similar to DFFITS, another diagnostic that is meant to show
    how influential a point is in a statistical regression. Although the computed values
    of Cook's and DFFITS are different, they are conceptually identical and there even
    exists a closed-form formula to convert one value to another. Because of this, we
    have chosen to implement Cook's distance rather than or in addition to DFFITS.
    """

    def __init__(
        self, ax=None, draw_threshold=True, linefmt="C0-", markerfmt=",", **kwargs
    ):
        # Initialize the visualizer
        super(CooksDistance, self).__init__(ax=ax, **kwargs)

        # Set "hyperparameters"
        self.draw_threshold = draw_threshold
        self.linefmt = linefmt
        self.markerfmt = markerfmt

        # An internal LinearRegression used to compute the residuals and MSE
        # This implementation doesn't support any regressor, it is OLS-specific
        self._model = LinearRegression()

    def fit(self, X, y):
        """
        Computes the leverage of X and uses the residuals of a
        ``sklearn.linear_model.LinearRegression`` to compute the Cook's Distance of each
        observation in X, their p-values and the number of outliers defined by the
        number of observations supplied.

        Parameters
        ----------
        X : array-like, 2D
            The exogenous design matrix, e.g. training data.

        y : array-like, 1D
            The endogenous response variable, e.g. target data.

        Returns
        -------
        self : CooksDistance
            Fit returns the visualizer instance.
        """
        # Fit a linear model to X and y to compute MSE
        self._model.fit(X, y)

        # Leverage is computed as the diagonal of the projection matrix of X
        # TODO: whiten X before computing leverage
        leverage = (X * np.linalg.pinv(X).T).sum(1)

        # Compute the rank and the degrees of freedom of the OLS model
        rank = np.linalg.matrix_rank(X)
        df = X.shape[0] - rank

        # Compute the MSE from the residuals
        residuals = y - self._model.predict(X)
        mse = np.dot(residuals, residuals) / df

        # Compute Cook's distance
        residuals_studentized = residuals / np.sqrt(mse) / np.sqrt(1 - leverage)
        self.distance_ = residuals_studentized ** 2 / X.shape[1]
        self.distance_ *= leverage / (1 - leverage)

        # Compute the p-values of Cook's Distance
        # TODO: honestly this was done because it was only in the statsmodels
        # implementation... I have no idea what this is or why its important.
        self.p_values_ = sp.stats.f.sf(self.distance_, X.shape[1], df)

        # Compute the influence threshold rule of thumb
        self.influence_threshold_ = 4 / X.shape[0]
        self.outlier_percentage_ = (
            sum(self.distance_ > self.influence_threshold_) / X.shape[0]
        )
        self.outlier_percentage_ *= 100.0

        self.draw()
        return self

    def draw(self):
        """
        Draws a stem plot where each stem is the Cook's Distance of the instance at the
        index specified by the x axis. Optionaly draws a threshold line.
        """
        # Draw a stem plot with the influence for each instance
        _, _, baseline = self.ax.stem(
            self.distance_, linefmt=self.linefmt, markerfmt=self.markerfmt,
            use_line_collection=True
        )

        # No padding on either side of the instance index
        self.ax.set_xlim(0, len(self.distance_))

        # Draw the threshold for most influential points
        if self.draw_threshold:
            label = r"{:0.2f}% > $I_t$ ($I_t=\frac {{4}} {{n}}$)".format(
                self.outlier_percentage_
            )
            self.ax.axhline(
                self.influence_threshold_,
                ls="--",
                label=label,
                c=baseline.get_color(),
                lw=baseline.get_linewidth(),
            )

        return self.ax

    def finalize(self):
        """
        Prepares the visualization for presentation and reporting.
        """
        # Set the title and axis labels
        self.set_title("Cook's Distance Outlier Detection")
        self.ax.set_xlabel("instance index")
        self.ax.set_ylabel("influence (I)")

        # Only add the legend if the influence threshold has been plotted
        if self.draw_threshold:
            self.ax.legend(loc="best", frameon=True)


def cooks_distance(
    X,
    y,
    ax=None,
    draw_threshold=True,
    linefmt="C0-",
    markerfmt=",",
    show=True,
    **kwargs
):
    """
    Cook's Distance is a measure of how influential an instance is to the computation of
    a regression, e.g. if the instance is removed would the estimated coeficients of the
    underlying model be substantially changed? Because of this, Cook's Distance is
    generally used to detect outliers in standard, OLS regression. In fact, a general
    rule of thumb is that D(i) > 4/n is a good threshold for determining highly
    influential points as outliers and this visualizer can report the percentage of data
    that is above that threshold.

    This implementation of Cook's Distance assumes Ordinary Least Squares regression,
    and therefore embeds a ``sklearn.linear_model.LinearRegression`` under the hood.
    Distance is computed via the non-whitened leverage of the projection matrix,
    computed inside of ``fit()``. The results of this visualizer are therefore similar
    to, but not as advanced, as a similar computation using statsmodels. Computing the
    influence for other regression models requires leave one out validation and can be
    expensive to compute.

    .. seealso::
        For a longer discussion on detecting outliers in regression and computing
        leverage and influence, see `linear regression in python, outliers/leverage
        detect <http://bit.ly/2If2fga>`_ by Huiming Song.

    Parameters
    ----------
    X : array-like, 2D
        The exogenous design matrix, e.g. training data.

    y : array-like, 1D
        The endogenous response variable, e.g. target data.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    draw_threshold : bool, default: True
        Draw a horizontal line at D(i) == 4/n to easily identify the most influential
        points on the final regression. This will also draw a legend that specifies the
        percentage of data points that are above the threshold.

    linefmt : str, default: 'C0-'
        A string defining the properties of the vertical lines of the stem plot, usually
        this will be a color or a color and a line style. The default is simply a solid
        line with the first color of the color cycle.

    markerfmt: str, default: ','
        A string defining the properties of the markers at the stem plot heads. The
        default is "pixel", e.g. basically no marker head at the top of the stem plot.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence the final
        visualization (e.g. size or title parameters).
    """
    viz = CooksDistance(
        ax=ax,
        draw_threshold=draw_threshold,
        linefmt=linefmt,
        markerfmt=markerfmt,
        **kwargs
    )
    viz.fit(X, y)

    # Draw the final visualization
    if show:
        viz.show()
    else:
        viz.finalize()

    # Return the visualizer
    return viz



# yellowbrick.target
# Implements visualizers related to the dependent (target) variable, y.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Thu Jul 19 08:57:05 2018 -0400
#
# ID: __init__.py [d742c57] benjamin@bengfort.com $

"""
Implements visualizers related to the dependent (target) variable, y. For
example, the ClassBalance visualizer shows how many of each class are
represented in the target. Other utilities include detection of sequential vs
discrete classes, binarization and thresholding visualization, and feature
correlation visualizations.
"""

##########################################################################
## Imports
##########################################################################

# Hoist visualizers into the top level of the target package
from .class_balance import ClassBalance, class_balance
from .binning import BalancedBinningReference, balanced_binning_reference
from .feature_correlation import FeatureCorrelation

# Alias the TargetType defined in yellowbrick.utils.target
from yellowbrick.utils.target import TargetType



# yellowbrick.text.correlation
# Implementation of word correlation for text visualization.
#
# Author:   Patrick Deziel
# Created:  Sun May 1 19:43:41 2022 -0600
#
# Copyright (C) 2022 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: correlation.py [b652fc9] deziel.patrick@gmail.com $

"""
Implementation of word correlation for text visualization.
"""


##########################################################################
## Imports
##########################################################################

import numpy as np

from sklearn.feature_extraction.text import CountVectorizer

from yellowbrick.style import find_text_color
from yellowbrick.text.base import TextVisualizer
from yellowbrick.style.palettes import color_sequence
from yellowbrick.exceptions import YellowbrickValueError

##########################################################################
## Word Correlation Plot Visualizer
##########################################################################

class WordCorrelationPlot(TextVisualizer):
    """
    Word correlation illustrates the extent to which words in a corpus appear in the
    same documents.

    WordCorrelationPlot visualizes the binary correlation between words across
    documents as a heatmap. The correlation is defined using the mean square
    contingency coefficient (phi-coefficient) between any two words m and n. The
    coefficient is a value between -1 and 1, inclusive. A value close to 1 or -1
    indicates strong positive or negative correlation between m and n, while a value
    close to 0 indicates little or no correlation. The constructor takes one required
    argument, which is the list of words or n-grams to be plotted.

    Parameters
    ----------
    words : list of str
        The list of words or n-grams to be plotted. The words must be present in the
        provided corpus on fit().

    ignore_case : bool, default: False
        If True, all words will be converted to lowercase before processing.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on.

    cmap : str or cmap, default: "RdYlBu"
        Colormap to use for the heatmap.

    colorbar : bool, default: True
        If True, a colorbar will be added to the heatmap.

    fontsize : int, default: None
        Font size to use for the labels on the axes.

    kwargs : dict
        Pass any additional keyword arguments to the super class.

    Attributes
    ----------
    self.doc_term_matrix_ : array of shape (n_docs, n_features)
        The computed sparse document-term matrix containing binary values indicating if
        a word is present in a document.

    self.num_docs_ : int
        The number of observed documents in the corpus.

    self.vocab_ : dict
        A dictionary mapping words to their indices in the document-term matrix.

    self.num_features_ : int
        The number of features (word labels) in the resulting plot.

    self.correlation_matrix_ : ndarray of shape (n_features, n_features)
        The computed matrix containing the phi-coefficients between all features.
    """
    def __init__(
        self,
        words,
        ignore_case=False,
        ax=None,
        cmap="RdYlBu",
        colorbar=True,
        fontsize=None,
        **kwargs
    ):
        super(WordCorrelationPlot, self).__init__(ax=ax, **kwargs)

        # Visual parameters
        self.fontsize = fontsize
        self.colorbar = colorbar
        self.cmap = color_sequence(cmap)

        # Fitting parameters
        self.ignore_case = ignore_case
        self.words = self._construct_terms(words, ignore_case)
        self.ngram_range = self._compute_ngram_range()

    def _construct_terms(self, words, ignore_case):
        """
        Constructs the list of terms to be plotted based on the provided words. This
        performs input checking and removes duplicates to produce a list of valid words
        for fitting.
        """
        # Remove surrounding whitespace
        terms = [word.strip() for word in words if len(word.strip()) > 0]

        if len(terms) == 0:
            raise YellowbrickValueError("Must provide at least one word to plot.")

        # Convert to lowercase if ignore_case is set
        if ignore_case:
            terms = [word.lower() for word in terms]

        # Sort and remove duplicates
        return sorted(set(terms))

    def _compute_ngram_range(self):
        """
        Computes the n-gram range to use for vectorization based on the provided words.
        This allows the user to specify multi-word terms for plotting.
        """
        ngrams = [len(word.split()) for word in self.words]
        return (min(ngrams), max(ngrams))

    def _compute_coefficient(self, m, n):
        """
        Computes the phi-coefficient for two words m and n, which is a correlation
        value between -1 and 1 inclusive.
        """
        m_col = self.doc_term_matrix_.getcol(self.vocab_[m])
        n_col = self.doc_term_matrix_.getcol(self.vocab_[n])
        both = m_col.multiply(n_col).sum()
        m_total = m_col.sum()
        n_total = n_col.sum()
        only_m = m_total - both
        only_n = n_total - both
        neither = self.num_docs_ - both - only_m - only_n
        return ((both * neither) - (only_m * only_n)) / np.sqrt(m_total * n_total * (self.num_docs_ - m_total) * (self.num_docs_ - n_total))

    def fit(self, X, y=None):
        """
        The fit method is the primary drawing input for the word correlation
        visualization.

        Parameters
        ----------
        X : list of str or generator
            Should be provided as a list of strings or a generator yielding strings
            that represent the documents in the corpus.
        
        y : None
            Labels are not used for the word correlation visualization.

        Returns
        -------
        self: instance
            Returns the instance of the transformer/visualizer.

        Attributes
        ----------
        self.doc_term_matrix_ : array of shape (n_docs, n_features)
            The computed sparse document-term matrix containing binary values
            indicating if a word is present in a document.

        self.num_docs_ : int
            The number of observed documents in the corpus.

        self.vocab_ : dict
            A dictionary mapping words to their indices in the document-term matrix.

        self.num_features_ : int
            The number of features (word labels) in the resulting plot.

        self.correlation_matrix_ : ndarray of shape (n_features, n_features)
            The computed matrix containing the phi-coefficients between all features.
        """

        # Instantiate the CountVectorizer
        vecs = CountVectorizer(
            vocabulary=self.words,
            lowercase=self.ignore_case,
            ngram_range=self.ngram_range,
            binary=True
        )

        # Get the binary document counts for the target words
        self.doc_term_matrix_ = vecs.fit_transform(X)
        self.num_docs_ = self.doc_term_matrix_.shape[0]
        self.vocab_ = vecs.vocabulary_

        # Verify that all target words exist in the corpus
        for word in self.words:
            if self.doc_term_matrix_.getcol(self.vocab_[word]).sum() == 0:
                raise YellowbrickValueError("Word '{}' does not exist in the corpus.".format(word))

        # Compute the phi-coefficient for each pair of words
        self.num_features_ = len(self.words)
        self.correlation_matrix_ = np.zeros((self.num_features_, self.num_features_))
        for i, m in enumerate(self.words):
            for j, n in enumerate(self.words):
                self.correlation_matrix_[i, j] = self._compute_coefficient(m, n)

        self.draw(X)
        return self

    def draw(self, X):
        """
        Called from the fit() method, this metod draws the heatmap on the figure using
        the computed correlation matrix.
        """

        # Use correlation matrix data for the heatmap
        wc_display = self.correlation_matrix_

        # Set up the dimensions of the pcolormesh
        X, Y = np.arange(self.num_features_ + 1), np.arange(self.num_features_ + 1)
        self.ax.set_ylim(bottom=0, top=wc_display.shape[0])
        self.ax.set_xlim(left=0, right=wc_display.shape[1])

        # Set the words as the tick labels on the plot. The Y-axis is sorted from top
        # to bottom, the X-axis is sorted from left to right.
        xticklabels = self.words
        yticklabels = self.words[::-1]
        ticks = np.arange(self.num_features_) + 0.5
        self.ax.set(xticks=ticks, yticks=ticks)
        self.ax.set_xticklabels(xticklabels, rotation="vertical", fontsize=self.fontsize)
        self.ax.set_yticklabels(yticklabels, fontsize=self.fontsize)

        # Flip the Y-axis values so that they match the sorted labels
        wc_display = np.flipud(wc_display)

        # Draw the labels in each heatmap cell
        for x in X[:-1]:
            for y in Y[:-1]:
                # Get the correlation value for the cell
                value = wc_display[x, y]
                svalue = "{:.2f}".format(value)

                # Get a compatible text color for the cell
                base_color = self.cmap(value / 2 + 0.5)
                text_color = find_text_color(base_color)

                # Draw the text at the center of the cell
                # Note: x and y coordinates are swapped to match the pcolormesh
                cx, cy = y + 0.5, x + 0.5
                self.ax.text(cx, cy, svalue, va="center", ha="center", color=text_color, fontsize=self.fontsize)

        # Draw the heatmap
        g = self.ax.pcolormesh(X, Y, wc_display, cmap=self.cmap, vmin=-1, vmax=1)

        # Add the color bar
        if self.colorbar:
            self.ax.figure.colorbar(g, ax=self.ax)

        return self.ax

    def finalize(self):
        """
        Prepares the figure for rendering by adding the title. This method is usually
        called from show() and not directly by the user.
        """
        self.set_title("Word Correlation Plot")
        self.fig.tight_layout()

##########################################################################
## Quick Method
##########################################################################

def word_correlation(
    words,
    corpus,
    ignore_case=True,
    ax=None,
    cmap="RdYlBu",
    show=True,
    colorbar=True,
    fontsize=None,
    **kwargs
):
    """Word Correlation

    Displays the binary correlation between the given words across the documents in a
    corpus.
    For a list of words with length n, this produces an n x n heatmap of correlation
    values in the range [-1, 1].

    Parameters
    ----------

    words : list of str
        The corpus words to display in the heatmap.
    corpus : list of str or generator
        The corpus as a list of documents or a generator yielding documents.

    ignore_case : bool, default: True
        If True, all words will be converted to lowercase before proessing.

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    cmap : str, default: "RdYlBu"
        Colormap to use for the heatmap.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    colorbar : bool, default: True
        If True, adds a colorbar to the figure.

    fontsize : int, default: None
        If not None, sets the font size of the labels.
    """
    # Instantiate the visualizer
    visualizer = WordCorrelationPlot(
        words=words,
        lowercase=ignore_case,
        ax=ax,
        cmap=cmap,
        colorbar=colorbar,
        fontsize=fontsize,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(corpus)

    # Draw the final visualization
    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer


# yellowbrick
# A suite of visual analysis and diagnostic tools for machine learning.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Created:  Wed May 18 10:46:33 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [0c5ba04] benjamin@bengfort.com $

"""
A suite of visual analysis and diagnostic tools to facilitate feature
selection, model selection, and parameter tuning for machine learning.
"""

##########################################################################
## Imports
##########################################################################

# Capture the original matplotlib rcParams
import matplotlib as mpl

_orig_rc_params = mpl.rcParams.copy()

# Import the version number at the top level
from .version import get_version, __version_info__

# Import the style management functions
from .style.rcmod import reset_defaults, reset_orig
from .style.rcmod import set_aesthetic, set_style, set_palette
from .style.palettes import color_palette, set_color_codes

# Import yellowbrick functionality to the top level
# TODO: review top-level functionality
from .anscombe import anscombe
from .datasaurus import datasaurus
from .classifier import ROCAUC, ClassBalance, ClassificationScoreVisualizer

# from .classifier import crplot, rocplot
# from .regressor import peplot, residuals_plot


##########################################################################
## Set default aesthetics
##########################################################################

set_aesthetic()  # NOTE: modifies mpl.rcParams


##########################################################################
## Package Version
##########################################################################

__version__ = get_version(short=True)



# yellowbrick.pipeline
# Implements a visual pipeline that subclasses Scikit-Learn pipelines.
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 21:41:06 2016 -0400
#
# Copyright (C) 2016 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: pipeline.py [1efae1f] benjamin@bengfort.com $

"""
Implements a visual pipeline that subclasses Scikit-Learn pipelines.
"""

##########################################################################
## Imports
##########################################################################

from os import path
from .base import Visualizer
from .utils.helpers import slugify
from sklearn.pipeline import Pipeline


##########################################################################
## Visual Pipeline
##########################################################################


class VisualPipeline(Pipeline):
    """Pipeline of transforms and visualizers with a final estimator.

    Sequentially apply a list of transforms, visualizers, and a final
    estimator which may be evaluated by additional visualizers.
    Intermediate steps of the pipeline must be 'transforms', that is, they
    must implement fit and transform methods.
    The final estimator only needs to implement fit.

    Any step that implements draw or show methods can be called sequentially
    directly from the VisualPipeline, allowing multiple visual diagnostics to
    be generated, displayed, and saved on demand.
    If draw or show is not called, the visual pipeline should be equivalent to
    the simple pipeline to ensure no reduction in performance.

    The purpose of the pipeline is to assemble several steps that can be
    cross-validated together while setting different parameters. These steps
    can be visually diagnosed by visualizers at every point in the pipeline.

    Parameters
    ----------
    steps : list
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator. Any intermediate step can be a FeatureVisualizer and the
        last step can be a ScoreVisualizer.

    Attributes
    ----------
    named_steps : dict
        Read-only attribute to access any step parameter by user given name.
        Keys are step names and values are step parameters.

    visual_steps : dict
        Read-only attribute to access any visualizer in he pipeline by user
        given name. Keys are step names and values are visualizer steps.

    """

    @property
    def visual_steps(self):
        return dict(step for step in self.steps if isinstance(step[1], Visualizer))

    def show(self, outdir=None, ext=".pdf", **kwargs):
        """
        A single entry point to rendering all visualizations in the visual
        pipeline. The rendering for the output depends on the backend context,
        but for path based renderings (e.g. saving to a file), specify a
        directory and extension to compse an outpath to save each
        visualization (file names will be based on the  named step).

        Parameters
        ----------
        outdir : path
            The directory to save visualizations to.

        ext : string, default = ".pdf"
            The extension of the file to save the visualization to.

        kwargs : dict
            Keyword arguments to pass to the ``show()`` method of all steps.
        """
        axes = []
        for name, step in self.visual_steps.items():
            if outdir is not None:
                outpath = path.join(outdir, slugify(name) + ext)
            else:
                outpath = None

            ax = step.show(outpath=outpath, **kwargs)
            axes.append(ax)

        # Return axes array to ensure figures are shown in notebook
        return axes

    def fit_transform_show(self, X, y=None, outpath=None, **kwargs):
        """
        Fit the model and transforms and then call show.
        """
        Xp = self.fit_transform(X, y, **kwargs)
        self.show(outpath, **kwargs)
        return Xp



# yellowbrick.classifier.rocauc
# Implements visual ROC/AUC curves for classification evaluation.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Neal Humphrey
# Created:  Tue May 03 18:15:42 2017 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: rocauc.py [5388065] neal@nhumphrey.com $

"""
Implements visual ROC/AUC curves for classification evaluation.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from sklearn.metrics import auc, roc_curve
from sklearn.preprocessing import label_binarize
from sklearn.utils.multiclass import type_of_target

from yellowbrick.exceptions import ModelError
from yellowbrick.style.palettes import LINE_COLOR
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.classifier.base import ClassificationScoreVisualizer


# Dictionary keys for ROCAUC
MACRO = "macro"
MICRO = "micro"

# Target Type Constants
BINARY = "binary"
MULTICLASS = "multiclass"


##########################################################################
## ROCAUC Visualizer
##########################################################################


class ROCAUC(ClassificationScoreVisualizer):
    """
    Receiver Operating Characteristic (ROC) curves are a measure of a
    classifier's predictive quality that compares and visualizes the tradeoff
    between the models' sensitivity and specificity. The ROC curve displays
    the true positive rate on the Y axis and the false positive rate on the
    X axis on both a global average and per-class basis. The ideal point is
    therefore the top-left corner of the plot: false positives are zero and
    true positives are one.

    This leads to another metric, area under the curve (AUC), a computation
    of the relationship between false positives and true positives. The higher
    the AUC, the better the model generally is. However, it is also important
    to inspect the "steepness" of the curve, as this describes the
    maximization of the true positive rate while minimizing the false positive
    rate. Generalizing "steepness" usually leads to discussions about
    convexity, which we do not get into here.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    micro : bool, default: True
        Plot the micro-averages ROC curve, computed from the sum of all true
        positives and false positives across all classes. Micro is not defined
        for binary classification problems with estimators with only a
        decision_function method.

    macro : bool, default: True
        Plot the macro-averages ROC curve, which simply takes the average of
        curves across all classes. Macro is not defined for binary
        classification problems with estimators with only a decision_function
        method.

    per_class : bool, default: True
        Plot the ROC curves for each individual class. This should be set
        to false if only the macro or micro average curves are required. For true
        binary classifiers, setting per_class=False will plot the positive class
        ROC curve, and per_class=True will use ``1-P(1)`` to compute the curve of
        the negative class if only a decision_function method exists on the estimator.

    binary : bool, default: False
        This argument quickly resets the visualizer for true binary classification
        by updating the micro, macro, and per_class arguments to False (do not use
        in conjunction with those other arguments). Note that this is not a true
        hyperparameter to the visualizer, it just collects other parameters into
        a single, simpler argument.

    classes : list of str, defult: None
        The class labels to use for the legend ordered by the index of the sorted
        classes discovered in the ``fit()`` method. Specifying classes in this
        manner is used to change the class names to a more specific format or
        to label encoded integer classes. Some visualizers may also use this
        field to filter the visualization for specific classes. For more advanced
        usage specify an encoder rather than class labels.

    encoder : dict or LabelEncoder, default: None
        A mapping of classes to human readable labels. Often there is a mismatch
        between desired class labels and those contained in the target variable
        passed to ``fit()`` or ``score()``. The encoder disambiguates this mismatch
        ensuring that classes are labeled correctly in the visualization.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The class labels observed while fitting.

    class_count_ : ndarray of shape (n_classes,)
        Number of samples encountered for each class during fitting.

    score_ : float
        An evaluation metric of the classifier on test data produced when
        ``score()`` is called. This metric is between 0 and 1 -- higher scores are
        generally better. For classifiers, this score is usually accuracy, but
        if micro or macro is specified this returns an F1 score.

    target_type_ : string
        Specifies if the detected classification target was binary or multiclass.

    Notes
    -----
    ROC curves are typically used in binary classification, and in fact the
    Scikit-Learn ``roc_curve`` metric is only able to perform metrics for
    binary classifiers. As a result it is necessary to binarize the output or
    to use one-vs-rest or one-vs-all strategies of classification. The
    visualizer does its best to handle multiple situations, but exceptions can
    arise from unexpected models or outputs.

    Another important point is the relationship of class labels specified on
    initialization to those drawn on the curves. The classes are not used to
    constrain ordering or filter curves; the ROC computation happens on the
    unique values specified in the target vector to the ``score`` method. To
    ensure the best quality visualization, do not use a LabelEncoder for this
    and do not pass in class labels.

    .. seealso::
        http://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html

    .. todo:: Allow the class list to filter the curves on the visualization.

    Examples
    --------
    >>> from yellowbrick.classifier import ROCAUC
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import train_test_split
    >>> data = load_data("occupancy")
    >>> features = ["temp", "relative humidity", "light", "C02", "humidity"]
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>> oz = ROCAUC(LogisticRegression())
    >>> oz.fit(X_train, y_train)
    >>> oz.score(X_test, y_test)
    >>> oz.show()
    """

    def __init__(
        self,
        estimator,
        ax=None,
        micro=True,
        macro=True,
        per_class=True,
        binary=False,
        classes=None,
        encoder=None,
        is_fitted="auto",
        force_model=False,
        **kwargs
    ):
        super(ROCAUC, self).__init__(
            estimator,
            ax=ax,
            classes=classes,
            encoder=encoder,
            is_fitted=is_fitted,
            force_model=force_model,
            **kwargs
        )

        # Set the visual parameters for ROCAUC
        # NOTE: the binary flag breaks our API since it's really just a meta parameter
        # for micro, macro, and per_class. We knew this going in, but did it anyway.
        
        self.binary = binary
        
        if self.binary:
            self.micro = False
            self.macro = False
            self.per_class = False
        else:
            self.micro = micro
            self.macro = macro
            self.per_class = per_class

    def fit(self, X, y=None):
        """
        Fit the classification model.
        """
        # The target determines what kind of estimator is fit
        ttype = type_of_target(y)
        if ttype.startswith(MULTICLASS):
            self.target_type_ = MULTICLASS
        elif ttype.startswith(BINARY):
            self.target_type_ = BINARY
        else:
            raise YellowbrickValueError(
                (
                    "{} does not support target type '{}', "
                    "please provide a binary or multiclass single-output target"
                ).format(self.__class__.__name__, ttype)
            )

        # Fit the model and return self
        return super(ROCAUC, self).fit(X, y)

    def score(self, X, y=None):
        """
        Generates the predicted target values using the Scikit-Learn
        estimator.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        Returns
        -------
        score_ : float
            Global accuracy unless micro or macro scores are requested.
        """
        # Call super to check if fitted and to compute self.score_
        # NOTE: this sets score to the base score if neither macro nor micro
        super(ROCAUC, self).score(X, y)

        # Compute the predictions for the test data
        y_pred = self._get_y_scores(X)

        if self.target_type_ == BINARY:
            # For binary, per_class must be True to draw micro/macro curves
            if (self.micro or self.macro) and not self.per_class:
                raise ModelError(
                    "no curves will be drawn; ",
                    "set per_class=True or micro=False and macro=False.",
                )

            # For binary, if predictions are returned in shape (n,), micro and macro
            # curves are not defined
            if (self.micro or self.macro) and len(y_pred.shape) == 1:
                raise ModelError(
                    "no curves will be drawn; set binary=True.",
                )

        if self.target_type_ == MULTICLASS:
            # If it's multiclass classification, at least one of micro, macro, or
            # per_class must be True
            if not self.micro and not self.macro and not self.per_class:
                raise YellowbrickValueError(
                    "no curves will be drawn; specify micro, macro, or per_class"
                )

        # Classes may be label encoded so only use what's in y to compute.
        # The self.classes_ attribute will be used as names for labels.
        classes = np.unique(y)
        n_classes = len(classes)

        # Store the false positive rate, true positive rate and curve info.
        self.fpr = dict()
        self.tpr = dict()
        self.roc_auc = dict()

        # If the decision is binary draw only ROC curve for the positive class
        if self.target_type_ is BINARY and not self.per_class:
            # In this case predict_proba returns an array of shape (n, 2) which
            # specifies the probabilities of both the negative and positive classes.
            if len(y_pred.shape) == 2 and y_pred.shape[1] == 2:
                self.fpr[BINARY], self.tpr[BINARY], _ = roc_curve(y, y_pred[:, 1])
            else:
                # decision_function returns array of shape (n,), so plot it directly
                self.fpr[BINARY], self.tpr[BINARY], _ = roc_curve(y, y_pred)
            self.roc_auc[BINARY] = auc(self.fpr[BINARY], self.tpr[BINARY])

        # Per-class binary decisions may have to have the negative class curve computed
        elif self.target_type_ is BINARY and self.per_class:
            # draw a curve for class 1 (the positive class)
            if len(y_pred.shape) == 2 and y_pred.shape[1] == 2:
                # predict_proba returns array of shape (n, 2), so use
                # probability of class 1 to compute ROC
                self.fpr[1], self.tpr[1], _ = roc_curve(y, y_pred[:, 1])
            else:
                # decision_function returns array of shape (n,)
                self.fpr[1], self.tpr[1], _ = roc_curve(y, y_pred)
            self.roc_auc[1] = auc(self.fpr[1], self.tpr[1])

            # draw a curve for class 0 (the negative class)
            if len(y_pred.shape) == 2 and y_pred.shape[1] == 2:
                # predict_proba returns array of shape (n, 2), so use
                # probability of class 0 to compute ROC
                self.fpr[0], self.tpr[0], _ = roc_curve(1 - y, y_pred[:, 0])
            else:
                # decision_function returns array of shape (n,).
                # To draw a ROC curve for class 0 we swap the classes 0 and 1 in y
                # and reverse classifiers predictions y_pred.
                self.fpr[0], self.tpr[0], _ = roc_curve(1 - y, -y_pred)
            self.roc_auc[0] = auc(self.fpr[0], self.tpr[0])

        else:
            # Otherwise compute the ROC curve and ROC area for each class
            for i, c in enumerate(classes):
                self.fpr[i], self.tpr[i], _ = roc_curve(y, y_pred[:, i], pos_label=c)
                self.roc_auc[i] = auc(self.fpr[i], self.tpr[i])

        # Compute micro average
        if self.micro:
            self._score_micro_average(y, y_pred, classes, n_classes)

        # Compute macro average
        if self.macro:
            self._score_macro_average(n_classes)

        # Draw the Curves
        self.draw()

        # Set score to micro average if specified
        if self.micro:
            self.score_ = self.roc_auc[MICRO]

        # Set score to macro average if not micro
        if self.macro:
            self.score_ = self.roc_auc[MACRO]

        return self.score_

    def draw(self):
        """
        Renders ROC-AUC plot.
        Called internally by score, possibly more than once

        Returns
        -------
        ax : the axis with the plotted figure
        """
        colors = self.class_colors_[0 : len(self.classes_)]
        n_classes = len(colors)

        # If it's a binary decision, plot the single ROC curve
        if self.target_type_ == BINARY and not self.per_class:
            self.ax.plot(
                self.fpr[BINARY],
                self.tpr[BINARY],
                label="ROC for binary decision, AUC = {:0.2f}".format(
                    self.roc_auc[BINARY]
                ),
            )

        # If per-class plotting is requested, plot ROC curves for each class
        if self.per_class:
            for i, color in zip(range(n_classes), colors):
                self.ax.plot(
                    self.fpr[i],
                    self.tpr[i],
                    color=color,
                    label="ROC of class {}, AUC = {:0.2f}".format(
                        self.classes_[i], self.roc_auc[i]
                    ),
                )

        # If requested, plot the ROC curve for the micro average
        if self.micro:
            self.ax.plot(
                self.fpr[MICRO],
                self.tpr[MICRO],
                linestyle="--",
                color=self.class_colors_[len(self.classes_) - 1],
                label="micro-average ROC curve, AUC = {:0.2f}".format(
                    self.roc_auc["micro"]
                ),
            )

        # If requested, plot the ROC curve for the macro average
        if self.macro:
            self.ax.plot(
                self.fpr[MACRO],
                self.tpr[MACRO],
                linestyle="--",
                color=self.class_colors_[len(self.classes_) - 1],
                label="macro-average ROC curve, AUC = {:0.2f}".format(
                    self.roc_auc["macro"]
                ),
            )

        # Plot the line of no discrimination to compare the curve to.
        self.ax.plot([0, 1], [0, 1], linestyle=":", c=LINE_COLOR)
        return self.ax

    def finalize(self, **kwargs):
        """
        Sets a title and axis labels of the figures and ensures the axis limits
        are scaled between the valid ROCAUC score values.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title and add the legend
        self.set_title("ROC Curves for {}".format(self.name))
        self.ax.legend(loc="lower right", frameon=True)

        # Set the limits for the ROC/AUC (always between 0 and 1)
        self.ax.set_xlim([0.0, 1.0])
        self.ax.set_ylim([0.0, 1.0])

        # Set x and y axis labels
        self.ax.set_ylabel("True Positive Rate")
        self.ax.set_xlabel("False Positive Rate")

    def _get_y_scores(self, X):
        """
        The ``roc_curve`` metric requires target scores that can either be the
        probability estimates of the positive class, confidence values or non-
        thresholded measure of decisions (as returned by "decision_function").

        This method computes the scores by resolving the estimator methods
        that retreive these values.

        .. todo:: implement confidence values metric.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features -- generally the test data
            that is associated with y_true values.
        """
        # The resolution order of scoring functions
        attrs = ("predict_proba", "decision_function")

        # Return the first resolved function
        for attr in attrs:
            try:
                method = getattr(self.estimator, attr, None)
                if method:
                    return method(X)
            except AttributeError:
                # Some Scikit-Learn estimators have both probability and
                # decision functions but override __getattr__ and raise an
                # AttributeError on access.
                # Note that because of the ordering of our attrs above,
                # estimators with both will *only* ever use probability.
                continue

        # If we've gotten this far, raise an error
        raise ModelError(
            "ROCAUC requires estimators with predict_proba or "
            "decision_function methods."
        )

    def _score_micro_average(self, y, y_pred, classes, n_classes):
        """
        Compute the micro average scores for the ROCAUC curves.
        """
        # Convert y to binarized array for micro and macro scores
        y = label_binarize(y, classes=classes)
        if n_classes == 2:
            y = np.hstack((1 - y, y))

        # Compute micro-average
        self.fpr[MICRO], self.tpr[MICRO], _ = roc_curve(y.ravel(), y_pred.ravel())
        self.roc_auc[MICRO] = auc(self.fpr[MICRO], self.tpr[MICRO])

    def _score_macro_average(self, n_classes):
        """
        Compute the macro average scores for the ROCAUC curves.
        """
        # Gather all FPRs
        all_fpr = np.unique(np.concatenate([self.fpr[i] for i in range(n_classes)]))
        avg_tpr = np.zeros_like(all_fpr)

        # Compute the averages per class
        for i in range(n_classes):
            avg_tpr += np.interp(all_fpr, self.fpr[i], self.tpr[i])

        # Finalize the average
        avg_tpr /= n_classes

        # Store the macro averages
        self.fpr[MACRO] = all_fpr
        self.tpr[MACRO] = avg_tpr
        self.roc_auc[MACRO] = auc(self.fpr[MACRO], self.tpr[MACRO])


##########################################################################
## Quick method for ROCAUC
##########################################################################


def roc_auc(
    estimator,
    X_train,
    y_train,
    X_test=None,
    y_test=None,
    ax=None,
    micro=True,
    macro=True,
    per_class=True,
    binary=False,
    classes=None,
    encoder=None,
    is_fitted="auto",
    force_model=False,
    show=True,
    **kwargs
):
    """ROCAUC

    Receiver Operating Characteristic (ROC) curves are a measure of a
    classifier's predictive quality that compares and visualizes the tradeoff
    between the models' sensitivity and specificity. The ROC curve displays
    the true positive rate on the Y axis and the false positive rate on the
    X axis on both a global average and per-class basis. The ideal point is
    therefore the top-left corner of the plot: false positives are zero and
    true positives are one.

    This leads to another metric, area under the curve  (AUC), a computation
    of the relationship between false positives and true positives. The higher
    the AUC, the better the model generally is. However, it is also important
    to inspect the "steepness" of the curve, as this describes the
    maximization of the true positive rate while minimizing the false positive
    rate. Generalizing "steepness" usually leads to discussions about
    convexity, which we do not get into here.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    X_train : array-like, 2D
        The table of instance data or independent variables that describe the outcome of
        the dependent variable, y. Used to fit the visualizer and also to score the
        visualizer if test splits are not specified.

    y_train : array-like, 2D
        The vector of target data or the dependent variable predicted by X. Used to fit
        the visualizer and also to score the visualizer if test splits not specified.

    X_test: array-like, 2D, default: None
        The table of instance data or independent variables that describe the outcome of
        the dependent variable, y. Used to score the visualizer if specified.

    y_test: array-like, 1D, default: None
        The vector of target data or the dependent variable predicted by X.
        Used to score the visualizer if specified.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    test_size : float, default=0.2
        The percentage of the data to reserve as test data.

    random_state : int or None, default=None
        The value to seed the random number generator for shuffling data.

    micro : bool, default: True
        Plot the micro-averages ROC curve, computed from the sum of all true
        positives and false positives across all classes. Micro is not defined
        for binary classification problems with estimators with only a
        decision_function method.

    macro : bool, default: True
        Plot the macro-averages ROC curve, which simply takes the average of
        curves across all classes. Macro is not defined for binary
        classification problems with estimators with only a decision_function
        method.

    per_class : bool, default: True
        Plot the ROC curves for each individual class. This should be set
        to false if only the macro or micro average curves are required. For true
        binary classifiers, setting per_class=False will plot the positive class
        ROC curve, and per_class=True will use ``1-P(1)`` to compute the curve of
        the negative class if only a decision_function method exists on the estimator.

    binary : bool, default: False
        This argument quickly resets the visualizer for true binary classification
        by updating the micro, macro, and per_class arguments to False (do not use
        in conjunction with those other arguments). Note that this is not a true
        hyperparameter to the visualizer, it just collects other parameters into
        a single, simpler argument.

    classes : list of str, defult: None
        The class labels to use for the legend ordered by the index of the sorted
        classes discovered in the ``fit()`` method. Specifying classes in this
        manner is used to change the class names to a more specific format or
        to label encoded integer classes. Some visualizers may also use this
        field to filter the visualization for specific classes. For more advanced
        usage specify an encoder rather than class labels.

    encoder : dict or LabelEncoder, default: None
        A mapping of classes to human readable labels. Often there is a mismatch
        between desired class labels and those contained in the target variable
        passed to ``fit()`` or ``score()``. The encoder disambiguates this mismatch
        ensuring that classes are labeled correctly in the visualization.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Notes
    -----
    ROC curves are typically used in binary classification, and in fact the
    Scikit-Learn ``roc_curve`` metric is only able to perform metrics for
    binary classifiers. As a result it is necessary to binarize the output or
    to use one-vs-rest or one-vs-all strategies of classification. The
    visualizer does its best to handle multiple situations, but exceptions can
    arise from unexpected models or outputs.

    Another important point is the relationship of class labels specified on
    initialization to those drawn on the curves. The classes are not used to
    constrain ordering or filter curves; the ROC computation happens on the
    unique values specified in the target vector to the ``score`` method. To
    ensure the best quality visualization, do not use a LabelEncoder for this
    and do not pass in class labels.

    .. seealso:: https://bit.ly/2IORWO2
    .. todo:: Allow the class list to filter the curves on the visualization.

    Examples
    --------
    >>> from yellowbrick.classifier import ROCAUC
    >>> from sklearn.linear_model import LogisticRegression
    >>> data = load_data("occupancy")
    >>> features = ["temp", "relative humidity", "light", "C02", "humidity"]
    >>> X = data[features].values
    >>> y = data.occupancy.values
    >>> roc_auc(LogisticRegression(), X, y)

    Returns
    -------
    viz : ROCAUC
        Returns the fitted, finalized visualizer object
    """
    # Instantiate the visualizer
    visualizer = ROCAUC(
        estimator=estimator,
        ax=ax,
        micro=micro,
        macro=macro,
        per_class=per_class,
        binary=binary,
        classes=classes,
        encoder=encoder,
        is_fitted=is_fitted,
        force_model=force_model,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X_train, y_train, **kwargs)

    # Scores the visualizer with X_test and y_test if provided,
    # X_train, y_train if not provided
    if X_test is not None and y_test is not None:
        visualizer.score(X_test, y_test)
    else:
        visualizer.score(X_train, y_train)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer



# yellowbrick.contrib.missing
# Visualizations related to missing values
#
# Author:  Nathan Danielsen
# Created: Fri Mar 29 5:17:36 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [1443e16] ndanielsen@users.noreply.github.com $

from .bar import MissingValuesBar, missing_bar
from .dispersion import MissingValuesDispersion, missing_dispersion



# yellowbrick.model_selection.rfecv
# Visualize the number of features selected with recursive feature elimination
#
# Author:  Benjamin Bengfort
# Created: Tue Apr 03 17:31:37 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: rfecv.py [a4599db] rebeccabilbro@users.noreply.github.com $

"""
Visualize the number of features selected using recursive feature elimination
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from yellowbrick.base import ModelVisualizer
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.utils import check_X_y
from sklearn.feature_selection import RFE
from sklearn.model_selection import cross_val_score


##########################################################################
## Recursive Feature Elimination
##########################################################################


class RFECV(ModelVisualizer):
    """
    Recursive Feature Elimination, Cross-Validated (RFECV) feature selection.

    Selects the best subset of features for the supplied estimator by removing
    0 to N features (where N is the number of features) using recursive
    feature elimination, then selecting the best subset based on the
    cross-validation score of the model. Recursive feature elimination
    eliminates n features from a model by fitting the model multiple times and
    at each step, removing the weakest features, determined by either the
    ``coef_`` or ``feature_importances_`` attribute of the fitted model.

    The visualization plots the score relative to each subset and shows trends
    in feature elimination. If the feature elimination CV score is flat, then
    potentially there are not enough features in the model. An ideal curve is
    when the score jumps from low to high as the number of features removed
    increases, then slowly decreases again from the optimal number of
    features.

    Parameters
    ----------
    estimator : a scikit-learn estimator
        An object that implements ``fit`` and provides information about the
        relative importance of features with either a ``coef_`` or
        ``feature_importances_`` attribute.

        Note that the object is cloned for each validation.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    step : int or float, optional (default=1)
        If greater than or equal to 1, then step corresponds to the (integer)
        number of features to remove at each iteration. If within (0.0, 1.0),
        then step corresponds to the percentage (rounded down) of features to
        remove at each iteration.

    groups : array-like, with shape (n_samples,), optional
        Group labels for the samples used while splitting the dataset into
        train/test set.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 3-fold cross-validation,
          - integer, to specify the number of folds.
          - An object to be used as a cross-validation generator.
          - An iterable yielding train/test splits.

        see the scikit-learn
        `cross-validation guide <http://scikit-learn.org/stable/modules/cross_validation.html>`_
        for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``. See scikit-learn model evaluation
        documentation for names of possible metrics.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    n_features_ : int
        The number of features in the selected subset

    support_ : array of shape [n_features]
        A mask of the selected features

    ranking_ : array of shape [n_features]
        The feature ranking, such that ``ranking_[i]`` corresponds to the
        ranked position of feature i. Selected features are assigned rank 1.

    cv_scores_ : array of shape [n_subsets_of_features, n_splits]
        The cross-validation scores for each subset of features and splits in
        the cross-validation strategy.

    rfe_estimator_ : sklearn.feature_selection.RFE
        A fitted RFE estimator wrapping the original estimator. All estimator
        functions such as ``predict()`` and ``score()`` are passed through to
        this estimator (it rewraps the original model).

    n_feature_subsets_ : array of shape [n_subsets_of_features]
        The number of features removed on each iteration of RFE, computed by the
        number of features in the dataset and the step parameter.

    Notes
    -----
    This model wraps ``sklearn.feature_selection.RFE`` and not
    ``sklearn.feature_selection.RFECV`` because access to the internals of the
    CV and RFE estimators is required for the visualization. The visualizer
    does take similar arguments, however it does not expose the same internal
    attributes.

    Additionally, the RFE model can be accessed via the ``rfe_estimator_``
    attribute. Once fitted, the visualizer acts as a wrapper for this
    estimator and not for the original model passed to the model. This way the
    visualizer model can be used to make predictions.

    .. caution:: This visualizer requires a model that has either a ``coef_``
        or ``feature_importances_`` attribute when fitted.
    """

    def __init__(
        self, estimator, ax=None, step=1, groups=None, cv=None, scoring=None, **kwargs
    ):

        # Initialize the model visualizer
        super(RFECV, self).__init__(estimator, ax=ax, **kwargs)

        # Set parameters
        self.step = step
        self.groups = groups
        self.cv = cv
        self.scoring = scoring

    def fit(self, X, y=None):
        """
        Fits the RFECV with the wrapped model to the specified data and draws
        the rfecv curve with the optimal number of features found.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression.

        Returns
        -------
        self : instance
            Returns the instance of the RFECV visualizer.
        """
        X, y = check_X_y(X, y, "csr")
        n_features = X.shape[1]

        # This check is kind of unnecessary since RFE will do it, but it's
        # nice to get it out of the way ASAP and raise a meaningful error.
        if 0.0 < self.step < 1.0:
            step = int(max(1, self.step * n_features))
        else:
            step = int(self.step)

        if step <= 0:
            raise YellowbrickValueError("step must be >0")

        # Create the RFE model
        rfe = RFE(self.estimator, step=step)
        self.n_feature_subsets_ = np.arange(1, n_features + step, step)

        # Create the cross validation params
        # TODO: handle random state
        cv_params = {key: self.get_params()[key] for key in ("groups", "cv", "scoring")}

        # Perform cross-validation for each feature subset
        scores = []
        for n_features_to_select in self.n_feature_subsets_:
            rfe.set_params(n_features_to_select=n_features_to_select)
            scores.append(cross_val_score(rfe, X, y, **cv_params))

        # Convert scores to array
        self.cv_scores_ = np.array(scores)

        # Find the best RFE model
        bestidx = self.cv_scores_.mean(axis=1).argmax()
        self.n_features_ = self.n_feature_subsets_[bestidx]

        # Fit the final RFE model for the number of features
        self.rfe_estimator_ = rfe
        self.rfe_estimator_.set_params(n_features_to_select=self.n_features_)
        self.rfe_estimator_.fit(X, y)

        # Rewrap the visualizer to use the rfe estimator
        self._wrapped = self.rfe_estimator_

        # Hoist the RFE params to the visualizer
        self.support_ = self.rfe_estimator_.support_
        self.ranking_ = self.rfe_estimator_.ranking_

        self.draw()
        return self

    def draw(self, **kwargs):
        """
        Renders the rfecv curve.
        """
        # Compute the curves
        x = self.n_feature_subsets_
        means = self.cv_scores_.mean(axis=1)
        sigmas = self.cv_scores_.std(axis=1)

        # Plot one standard deviation above and below the mean
        self.ax.fill_between(x, means - sigmas, means + sigmas, alpha=0.25)

        # Plot the curve
        self.ax.plot(x, means, "o-")

        # Plot the maximum number of features
        self.ax.axvline(
            self.n_features_,
            c="k",
            ls="--",
            label="n_features = {}\nscore = {:0.3f}".format(
                self.n_features_, self.cv_scores_.mean(axis=1).max()
            ),
        )

        return self.ax

    def finalize(self, **kwargs):
        """
        Add the title, legend, and other visual final touches to the plot.
        """
        # Set the title of the figure
        self.set_title("RFECV for {}".format(self.name))

        # Add the legend
        self.ax.legend(frameon=True, loc="best")

        # Set the axis labels
        self.ax.set_xlabel("Number of Features Selected")
        self.ax.set_ylabel("Score")


##########################################################################
## Quick Methods
##########################################################################


def rfecv(
    estimator,
    X,
    y,
    ax=None,
    step=1,
    groups=None,
    cv=None,
    scoring=None,
    show=True,
    **kwargs
):
    """
    Performs recursive feature elimination with cross-validation to determine
    an optimal number of features for a model. Visualizes the feature subsets
    with respect to the cross-validation score.

    This helper function is a quick wrapper to utilize the RFECV visualizer
    for one-off analysis.

    Parameters
    ----------
    estimator : a scikit-learn estimator
        An object that implements ``fit`` and provides information about the
        relative importance of features with either a ``coef_`` or
        ``feature_importances_`` attribute.

        Note that the object is cloned for each validation.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    step : int or float, optional (default=1)
        If greater than or equal to 1, then step corresponds to the (integer)
        number of features to remove at each iteration. If within (0.0, 1.0),
        then step corresponds to the percentage (rounded down) of features to
        remove at each iteration.

    groups : array-like, with shape (n_samples,), optional
        Group labels for the samples used while splitting the dataset into
        train/test set.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 3-fold cross-validation,
          - integer, to specify the number of folds.
          - An object to be used as a cross-validation generator.
          - An iterable yielding train/test splits.

        see the scikit-learn
        `cross-validation guide <http://scikit-learn.org/stable/modules/cross_validation.html>`_
        for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``. See scikit-learn model evaluation
        documentation for names of possible metrics.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers. These arguments are
        also passed to the `show()` method, e.g. can pass a path to save the
        figure to.

    Returns
    -------
    viz : RFECV
        Returns the fitted, finalized visualizer.
    """
    # Initialize the visualizer
    oz = RFECV(
        estimator, ax=ax, step=step, groups=groups, cv=cv, scoring=scoring, show=show
    )

    # Fit and show the visualizer
    oz.fit(X, y)

    if show:
        oz.show()
    else:
        oz.finalize()

    # Return the visualizer object
    return oz



# yellowbrick.utils.wrapper
# Utility package that provides a wrapper for new style classes.
#
# Author:   Benjamin Bengfort
# Created:  Sun May 21 20:27:32 2017 -0700
#
# Copyright (C) 2017 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: wrapper.py [b2ecd50] benjamin@bengfort.com $

"""
Utility package that provides a wrapper for new style classes.
"""

##########################################################################
## Wrapper Class
##########################################################################

from yellowbrick.exceptions import YellowbrickAttributeError, YellowbrickTypeError


class Wrapper(object):
    """
    Object wrapper class.

    An object wrapper is initialized with the object it wraps then proxies
    any unhandled getattribute methods to it. If no attribute is found either
    on the wrapper or the wrapped object, an AttributeError is raised.

    .. seealso:: http://code.activestate.com/recipes/577555-object-wrapper-class/

    Parameters
    ----------
    obj : object
        The object to wrap with the wrapper class
    """

    def __init__(self, obj):
        self._wrapped = obj

    def __getattr__(self, attr):
        if self is self._wrapped:
            raise YellowbrickTypeError("wrapper cannot wrap itself or recursion will occur")

        # proxy to the wrapped object
        try:
            return getattr(self._wrapped, attr)
        except AttributeError as e:
            raise YellowbrickAttributeError(f"neither visualizer '{self.__class__.__name__}' nor wrapped estimator '{type(self._wrapped).__name__}' have attribute '{attr}'") from e



# yellowbrick.draw
# Utilities for common matplotlib drawing procedures.
#
# Author:  Benjamin Bengfort
# Created: Sun Aug 19 10:35:50 2018 -0400
#
# Copyright (C) 2018 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: draw.py [dd915ad] benjamin@bengfort.com $

"""
Utilities for common matplotlib drawing procedures.
"""

##########################################################################
## Imports
##########################################################################

from .base import Visualizer
from .exceptions import YellowbrickValueError
from .style.colors import resolve_colors

from matplotlib import patches

import matplotlib.pyplot as plt
import numpy as np

##########################################################################
## Legend Drawing Utilities
##########################################################################


def manual_legend(g, labels, colors, **legend_kwargs):
    """
    Adds a manual legend for a scatter plot to the visualizer where the labels
    and associated colors are drawn with circle patches instead of determining
    them from the labels of the artist objects on the axes. This helper is
    used either when there are a lot of duplicate labels, no labeled artists,
    or when the color of the legend doesn't exactly match the color in the
    figure (e.g. because of the use of transparency).

    Parameters
    ----------
    g : Visualizer or Axes object
        The graph to draw the legend on, either a Visualizer or a matplotlib
        Axes object. If None, the current axes are drawn on, but this is not
        recommended.

    labels : list of str
        The text labels to associate with the legend. Note that the labels
        will be added to the legend in the order specified.

    colors : list of colors
        A list of any valid matplotlib color reference. The number of colors
        specified must be equal to the number of labels.

    legend_kwargs : dict
        Any additional keyword arguments to pass to the legend.

    Returns
    -------
    legend: Legend artist
        The artist created by the ax.legend() call, returned for further
        manipulation if required by the caller.

    Notes
    -----
    Right now this method simply draws the patches as rectangles and cannot
    take into account the line or scatter plot properties (e.g. line style or
    marker style). It is possible to add Line2D patches to the artist that do
    add manual styles like this, which we can explore in the future.

    .. seealso:: https://matplotlib.org/gallery/text_labels_and_annotations/custom_legends.html
    """
    # Get access to the matplotlib Axes
    if isinstance(g, Visualizer):
        g = g.ax
    elif g is None:
        g = plt.gca()

    # Ensure that labels and colors are the same length to prevent odd behavior.
    if len(colors) != len(labels):
        raise YellowbrickValueError(
            "please specify the same number of colors as labels!"
        )

    # Create the legend handles with the associated colors and labels
    handles = [
        patches.Patch(color=color, label=label) for color, label in zip(colors, labels)
    ]

    # Return the Legend artist
    return g.legend(handles=handles, **legend_kwargs)


def bar_stack(
    data,
    ax=None,
    labels=None,
    ticks=None,
    colors=None,
    colormap=None,
    orientation="vertical",
    legend=True,
    legend_kws=None,
    **kwargs
):
    """
    An advanced bar chart plotting utility that can draw bar and stacked bar charts from
    data, wrapping calls to the specified matplotlib.Axes object.

    Parameters
    ----------
    data : 2D array-like
        The data passed to the Visualizer. Rows represent each stack in the bar chart and columns
        represent each bar. Therefore, a single bar chart is created by passing a 2D array
        containing a single row, while the data to create a bar chart with 3 stacks would have a
        shape of (3, b).

    ax : matplotlib.Axes, default: None
        The axes object to draw the barplot on, uses plt.gca() if not specified.

    labels : list of str, default: None
        The labels for each row in the bar stack, used to create a legend.

    ticks : list of str, default: None
        The labels for each bar, added to the x-axis for a vertical plot, or the y-axis
        for a horizontal plot.

    colors : array-like, default: None
        Specify the colors of each bar, each row in the stack, or every segment.

    colormap : string or matplotlib cmap
        Specify a colormap for each bar, each row in the stack, or every segment.

    orientation:‘vertical’ or ‘horizontal’
        Specifies a horizontal or vertical bar chart.

    legend : boolean, default: True
        If True, the function add a legend with the plot

    legend_kws : dict, default: None
        Additional keyword arguments for the legend components.

    kwargs : dict
        Additional keyword arguments to pass to ``ax.bar``.
    """
    if ax is None:
        ax = plt.gca()

    colors = resolve_colors(n_colors=data.shape[0], colormap=colormap, colors=colors)

    idx = np.arange(data.shape[1])
    zeros = np.zeros(data.shape[1])
    # Stores stacks for both side of plotting axes
    stack_arr = np.zeros((data.shape[1], 2))
    orientation = orientation.lower()

    if orientation.startswith("h"):

        for rdx in range(len(data)):
            stack = [stack_arr[j, int(data[rdx][j] > 0)] for j in range(len(data[rdx]))]
            ax.barh(idx, data[rdx], left=stack, color=colors[rdx])
            # Updates the stack for negative side of y-axis
            stack_arr[:, 0] += np.minimum(data[rdx], zeros)
            # Updates stack for positive side of y-axis
            stack_arr[:, 1] += np.maximum(data[rdx], zeros)
        ax.set_yticks(idx)
        if ticks is not None:
            ax.set_yticklabels(ticks)

    elif orientation.startswith("v"):
        for rdx in range(len(data)):
            stack = [stack_arr[j, int(data[rdx][j] > 0)] for j in range(len(data[rdx]))]
            ax.bar(idx, data[rdx], bottom=stack, color=colors[rdx])
            # Updates the stack for negative side of x-axis
            stack_arr[:, 0] += np.minimum(data[rdx], zeros)
            # Updates the stack for negative side of x-axis
            stack_arr[:, 1] += np.maximum(data[rdx], zeros)
        ax.set_xticks(idx)
        if ticks is not None:
            ax.set_xticklabels(ticks, rotation=90)

    else:
        raise YellowbrickValueError("unknown orientation '{}'".format(orientation))

    # Generates default labels is labels are not specified.
    labels = labels or np.arange(data.shape[0])

    if legend:
        legend_kws = legend_kws or {}
        manual_legend(ax, labels=labels, colors=colors, **legend_kws)
    return ax



# yellowbrick.style.rcmod
# Modifies the matplotlib rcParams in order to make yellowbrick appealing.
#
# Author:   Benjamin Bengfort
# Created:  Thu Oct 06 08:45:38 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: rcmod.py [c6aff34] benjamin@bengfort.com $

"""
Modifies the matplotlib rcParams in order to make yellowbrick more appealing.
This has been modified from Seaborn's rcmod.py: github.com/mwaskom/seaborn in
order to alter the matplotlib rc dictionary on the fly.

NOTE: matplotlib 2.0 styles mean we can simply convert this to a stylesheet!
"""

##########################################################################
## Imports
##########################################################################

import functools
import numpy as np
import matplotlib as mpl

# Check to see if we have a slightly modern version of mpl
from distutils.version import LooseVersion

mpl_ge_150 = LooseVersion(mpl.__version__) >= "1.5.0"


from .. import _orig_rc_params
from .palettes import color_palette, set_color_codes


##########################################################################
## Exports
##########################################################################

__all__ = ["set_aesthetic", "set_style", "set_palette", "reset_defaults", "reset_orig"]


##########################################################################
## rcParams Keys
##########################################################################

_style_keys = (
    "axes.facecolor",
    "axes.edgecolor",
    "axes.grid",
    "axes.axisbelow",
    "axes.linewidth",
    "axes.labelcolor",
    "figure.facecolor",
    "grid.color",
    "grid.linestyle",
    "text.color",
    "xtick.color",
    "ytick.color",
    "xtick.direction",
    "ytick.direction",
    "xtick.major.size",
    "ytick.major.size",
    "xtick.minor.size",
    "ytick.minor.size",
    "legend.frameon",
    "legend.numpoints",
    "legend.scatterpoints",
    "lines.solid_capstyle",
    "image.cmap",
    "font.family",
    "font.sans-serif",
)

_context_keys = (
    "figure.figsize",
    "font.size",
    "axes.labelsize",
    "axes.titlesize",
    "xtick.labelsize",
    "ytick.labelsize",
    "legend.fontsize",
    "grid.linewidth",
    "lines.linewidth",
    "patch.linewidth",
    "lines.markersize",
    "lines.markeredgewidth",
    "xtick.major.width",
    "ytick.major.width",
    "xtick.minor.width",
    "ytick.minor.width",
    "xtick.major.pad",
    "ytick.major.pad",
)


##########################################################################
## rcParams Keys
##########################################################################


def set_aesthetic(
    palette="yellowbrick", font="sans-serif", font_scale=1, color_codes=True, rc=None
):
    """
    Set aesthetic parameters in one step.

    Each set of parameters can be set directly or temporarily, see the
    referenced functions below for more information.

    Parameters
    ----------
    palette : string or sequence
        Color palette, see :func:`color_palette`
    font : string
        Font family, see matplotlib font manager.
    font_scale : float, optional
        Separate scaling factor to independently scale the size of the
        font elements.
    color_codes : bool
        If ``True`` and ``palette`` is a yellowbrick palette, remap the shorthand
        color codes (e.g. "b", "g", "r", etc.) to the colors from this palette.
    rc : dict or None
        Dictionary of rc parameter mappings to override the above.
    """
    _set_context(font_scale)
    set_style(rc={"font.family": font})
    set_palette(palette, color_codes=color_codes)
    if rc is not None:
        mpl.rcParams.update(rc)


def reset_defaults():
    """
    Restore all RC params to default settings.
    """
    mpl.rcParams.update(mpl.rcParamsDefault)


def reset_orig():
    """
    Restore all RC params to original settings (respects custom rc).
    """
    mpl.rcParams.update(_orig_rc_params)


##########################################################################
## Axes Styles
##########################################################################


def _axes_style(style=None, rc=None):
    """
    Return a parameter dict for the aesthetic style of the plots.

    NOTE: This is an internal method from Seaborn that is simply used to
    create a default aesthetic in yellowbrick. If you'd like to use these
    styles then import Seaborn!

    This affects things like the color of the axes, whether a grid is
    enabled by default, and other aesthetic elements.

    This function returns an object that can be used in a ``with`` statement
    to temporarily change the style parameters.

    Parameters
    ----------
    style : dict, reset, or None
        A dictionary of parameters or the name of a preconfigured set.

    rc : dict, optional
        Parameter mappings to override the values in the preset seaborn
        style dictionaries. This only updates parameters that are
        considered part of the style definition.

    """
    if isinstance(style, dict):
        style_dict = style

    else:
        # Define colors here
        dark_gray = ".15"
        light_gray = ".8"

        # Common parameters
        style_dict = {
            "figure.facecolor": "white",
            "text.color": dark_gray,
            "axes.labelcolor": dark_gray,
            "legend.frameon": False,
            "legend.numpoints": 1,
            "legend.scatterpoints": 1,
            "xtick.direction": "out",
            "ytick.direction": "out",
            "xtick.color": dark_gray,
            "ytick.color": dark_gray,
            "axes.axisbelow": True,
            "image.cmap": "Greys",
            "font.family": ["sans-serif"],
            "font.sans-serif": [
                "Arial",
                "Liberation Sans",
                "Bitstream Vera Sans",
                "sans-serif",
            ],
            "grid.linestyle": "-",
            "axes.grid": True,
            "lines.solid_capstyle": "round",
            "axes.facecolor": "white",
            "axes.edgecolor": light_gray,
            "axes.linewidth": 1.25,
            "grid.color": light_gray,
            "xtick.major.size": 0,
            "ytick.major.size": 0,
            "xtick.minor.size": 0,
            "ytick.minor.size": 0,
        }

    # Override these settings with the provided rc dictionary
    if rc is not None:
        rc = {k: v for k, v in rc.items() if k in _style_keys}
        style_dict.update(rc)

    # Wrap in an _AxesStyle object so this can be used in a with statement
    style_object = _AxesStyle(style_dict)

    return style_object


def set_style(style=None, rc=None):
    """
    Set the aesthetic style of the plots.

    This affects things like the color of the axes, whether a grid is
    enabled by default, and other aesthetic elements.

    Parameters
    ----------
    style : dict, None, or one of {darkgrid, whitegrid, dark, white, ticks}
        A dictionary of parameters or the name of a preconfigured set.
    rc : dict, optional
        Parameter mappings to override the values in the preset seaborn
        style dictionaries. This only updates parameters that are
        considered part of the style definition.
    """
    style_object = _axes_style(style, rc)
    mpl.rcParams.update(style_object)


##########################################################################
## Context
##########################################################################


def _plotting_context(context=None, font_scale=1, rc=None):
    """
    Return a parameter dict to scale elements of the figure.

    NOTE: This is an internal method from Seaborn that is simply used to
    create a default aesthetic in yellowbrick. If you'd like to use these
    styles then import Seaborn!

    This affects things like the size of the labels, lines, and other
    elements of the plot, but not the overall style. The base context
    is "notebook", and the other contexts are "paper", "talk", and "poster",
    which are version of the notebook parameters scaled by .8, 1.3, and 1.6,
    respectively.

    This function returns an object that can be used in a ``with`` statement
    to temporarily change the context parameters.

    Parameters
    ----------
    context : dict, None, or one of {paper, notebook, talk, poster}
        A dictionary of parameters or the name of a preconfigured set.
    font_scale : float, optional
        Separate scaling factor to independently scale the size of the
        font elements.
    rc : dict, optional
        Parameter mappings to override the values in the preset seaborn
        context dictionaries. This only updates parameters that are
        considered part of the context definition.

    """
    if isinstance(context, dict):
        context_dict = context

    else:
        # Set up dictionary of default parameters
        base_context = {
            "figure.figsize": np.array([8, 5.5]),
            "font.size": 12,
            "axes.labelsize": 11,
            "axes.titlesize": 12,
            "xtick.labelsize": 10,
            "ytick.labelsize": 10,
            "legend.fontsize": 10,
            "grid.linewidth": 1,
            "lines.linewidth": 1.75,
            "patch.linewidth": 0.3,
            "lines.markersize": 7,
            "lines.markeredgewidth": 0,
            "xtick.major.width": 1,
            "ytick.major.width": 1,
            "xtick.minor.width": 0.5,
            "ytick.minor.width": 0.5,
            "xtick.major.pad": 7,
            "ytick.major.pad": 7,
        }

        # Scale all the parameters by the same factor depending on the context
        scaling = dict(paper=0.8, notebook=1, talk=1.3, poster=1.6)["notebook"]
        context_dict = {k: v * scaling for k, v in base_context.items()}

        # Now independently scale the fonts
        font_keys = [
            "axes.labelsize",
            "axes.titlesize",
            "legend.fontsize",
            "xtick.labelsize",
            "ytick.labelsize",
            "font.size",
        ]
        font_dict = {k: context_dict[k] * font_scale for k in font_keys}
        context_dict.update(font_dict)

    # Implement hack workaround for matplotlib bug
    # See https://github.com/mwaskom/seaborn/issues/344
    # There is a bug in matplotlib 1.4.2 that makes points invisible when
    # they don't have an edgewidth. It will supposedly be fixed in 1.4.3.
    if mpl.__version__ == "1.4.2":
        context_dict["lines.markeredgewidth"] = 0.01

    # Override these settings with the provided rc dictionary
    if rc is not None:
        rc = {k: v for k, v in rc.items() if k in _context_keys}
        context_dict.update(rc)

    # Wrap in a _PlottingContext object so this can be used in a with statement
    context_object = _PlottingContext(context_dict)

    return context_object


def _set_context(context=None, font_scale=1, rc=None):
    """
    Set the plotting context parameters.

    NOTE: This is an internal method from Seaborn that is simply used to
    create a default aesthetic in yellowbrick. If you'd like to use these
    styles then import Seaborn!

    This affects things like the size of the labels, lines, and other
    elements of the plot, but not the overall style. The base context
    is "notebook", and the other contexts are "paper", "talk", and "poster",
    which are version of the notebook parameters scaled by .8, 1.3, and 1.6,
    respectively.

    Parameters
    ----------
    context : dict, None, or one of {paper, notebook, talk, poster}
        A dictionary of parameters or the name of a preconfigured set.
    font_scale : float, optional
        Separate scaling factor to independently scale the size of the
        font elements.
    rc : dict, optional
        Parameter mappings to override the values in the preset seaborn
        context dictionaries. This only updates parameters that are
        considered part of the context definition.

    """
    context_object = _plotting_context(context, font_scale, rc)
    mpl.rcParams.update(context_object)


class _RCAesthetics(dict):
    def __enter__(self):
        rc = mpl.rcParams
        self._orig = {k: rc[k] for k in self._keys}
        self._set(self)

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._set(self._orig)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper


class _AxesStyle(_RCAesthetics):
    """Light wrapper on a dict to set style temporarily."""

    _keys = _style_keys
    _set = staticmethod(set_style)


class _PlottingContext(_RCAesthetics):
    """Light wrapper on a dict to set context temporarily."""

    _keys = _context_keys
    _set = staticmethod(_set_context)


##########################################################################
## Colors/Palettes
##########################################################################


def set_palette(palette, n_colors=None, color_codes=False):
    """
    Set the matplotlib color cycle using a seaborn palette.

    Parameters
    ----------
    palette : yellowbrick color palette | seaborn color palette (with ``sns_`` prepended)
        Palette definition. Should be something that :func:`color_palette`
        can process.
    n_colors : int
        Number of colors in the cycle. The default number of colors will depend
        on the format of ``palette``, see the :func:`color_palette`
        documentation for more information.
    color_codes : bool
        If ``True`` and ``palette`` is a seaborn palette, remap the shorthand
        color codes (e.g. "b", "g", "r", etc.) to the colors from this palette.
    """
    colors = color_palette(palette, n_colors)
    if mpl_ge_150:
        from cycler import cycler

        cyl = cycler("color", colors)
        mpl.rcParams["axes.prop_cycle"] = cyl
    else:
        mpl.rcParams["axes.color_cycle"] = list(colors)
    mpl.rcParams["patch.facecolor"] = colors[0]
    if color_codes:
        set_color_codes(palette)



# yellowbrick.contrib.classifier
# Visualizations related to evaluating Scikit-Learn classification models
#
# Author:   Nathan Danielsen
# Created:  Wed Mar 29 12:39:40 2018 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [a60bc41] nathan.danielsen@gmail.com $

from .boundaries import decisionviz, DecisionBoundariesVisualizer, DecisionViz



# yellowbrick.colors
# Colors and color helpers brought in from a different library.
#
# Author:   Benjamin Bengfort
# Created:  Fri Jun 24 17:02:53 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: colors.py [c6aff34] benjamin@bengfort.com $

"""
Colors and color helpers brought in from an alternate library.
See https://bl.ocks.org/mbostock/5577023
"""

##########################################################################
## Imports
##########################################################################

import random
import warnings
import numpy as np
import matplotlib as mpl
import matplotlib.cm as cm

from copy import copy

from yellowbrick.exceptions import YellowbrickValueError


# Check to see if matplotlib is at least sorta up to date
from distutils.version import LooseVersion

mpl_ge_150 = LooseVersion(mpl.__version__) >= "1.5.0"


##########################################################################
## Color Utilities
##########################################################################


def get_color_cycle():
    """
    Returns the current color cycle from matplotlib.
    """
    if mpl_ge_150:
        cyl = mpl.rcParams["axes.prop_cycle"]
        # matplotlib 1.5 verifies that axes.prop_cycle *is* a cycler
        # but no garuantee that there's a `color` key.
        # so users could have a custom rcParams w/ no color...
        try:
            return [x["color"] for x in cyl]
        except KeyError:
            pass  # just return axes.color style below
    return mpl.rcParams["axes.color_cycle"]


def resolve_colors(n_colors=None, colormap=None, colors=None):
    """
    Generates a list of colors based on common color arguments, for example
    the name of a colormap or palette or another iterable of colors. The list
    is then truncated (or multiplied) to the specific number of requested
    colors.

    Parameters
    ----------
    n_colors : int, default: None
        Specify the length of the list of returned colors, which will either
        truncate or multiple the colors available. If None the length of the
        colors will not be modified.

    colormap : str, yellowbrick.style.palettes.ColorPalette, matplotlib.cm, default: None
        The name of the matplotlib color map with which to generate colors.

    colors : iterable, default: None
        A collection of colors to use specifically with the plot. Overrides
        colormap if both are specified.

    Returns
    -------
    colors : list
        A list of colors that can be used in matplotlib plots.

    Notes
    -----
    This function was originally based on a similar function in the pandas
    plotting library that has been removed in the new version of the library.
    """

    # Work with the colormap if specified and colors is not
    if colormap is not None and colors is None:
        # Must import here to avoid recursive import
        from .palettes import PALETTES, ColorPalette

        if isinstance(colormap, str):
            try:

                # try to get colormap from PALETTES first
                _colormap = PALETTES.get(colormap, None)

                if _colormap is None:

                    colormap = cm.get_cmap(colormap)
                    n_colors = n_colors or len(get_color_cycle())
                    _colors = list(map(colormap, np.linspace(0, 1, num=n_colors)))

                else:

                    _colors = ColorPalette(_colormap).as_rgb()
                    n_colors = n_colors or len(_colors)

            except ValueError as e:

                raise YellowbrickValueError(e)

        # if yellowbrick color palette is provided as colormap
        elif isinstance(colormap, ColorPalette):

            _colors = colormap.as_rgb()
            n_colors = n_colors or len(_colors)

        # if matplotlib color palette is provided as colormap
        elif isinstance(colormap, mpl.colors.Colormap):
            n_colors = n_colors or len(get_color_cycle())
            _colors = list(map(colormap, np.linspace(0, 1, num=n_colors)))
        else:
            raise YellowbrickValueError(
                "Colormap type {} is not recognized. Possible types are: {}".format(
                    type(colormap),
                    ", ".join(
                        ["yellowbrick.style.ColorPalette,", "matplotlib.cm,", "str"]
                    ),
                )
            )

    # Work with the color list
    elif colors is not None:

        # Warn if both colormap and colors is specified.
        if colormap is not None:
            warnings.warn("both colormap and colors specified; using colors")

        _colors = list(colors)  # Ensure colors is a list

    # Get the default colors
    else:
        _colors = get_color_cycle()

    # Truncate or multiple the color list according to the number of colors
    if n_colors is not None and len(_colors) != n_colors:
        _colors = [_colors[idx % len(_colors)] for idx in np.arange(n_colors)]

    return _colors


class ColorMap(object):
    """
    A helper for mapping categorical values to colors on demand.
    """

    def __init__(self, colors="flatui", shuffle=False):
        """
        Specify either a list of colors or one of the color names. If shuffle
        is True then the colors will be shuffled randomly.
        """
        self.mapping = {}
        self.colors = colors

        if shuffle:
            random.shuffle(self._colors)

    @property
    def colors(self):
        return self._colors

    @colors.setter
    def colors(self, value):
        """
        Converts color strings into a color listing.
        """
        if isinstance(value, str):
            # Must import here to avoid recursive import
            from .palettes import PALETTES

            if value not in PALETTES:
                raise YellowbrickValueError(
                    "'{}' is not a registered color palette".format(value)
                )
            self._colors = copy(PALETTES[value])
        elif isinstance(value, list):
            self._colors = value
        else:
            self._colors = list(value)

    def __call__(self, category):
        if category not in self.mapping:
            if self.colors:
                self.mapping[category] = self.colors.pop()
            else:
                raise YellowbrickValueError(
                    "Not enough colors for this many categories!"
                )

        return self.mapping[category]



# yellowbrick.model_selection.learning_curve
# Implements a learning curve visualization for model selection.
#
# Author:   Jason Keung
# Created:  Mon May 22 09:22:00 2017 -0500
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: learning_curve.py [c5355ee] benjamin@bengfort.com $

"""
Implements a learning curve visualization for model selection.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from yellowbrick.base import ModelVisualizer
from yellowbrick.style import resolve_colors
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.model_selection import learning_curve as sk_learning_curve


# Default ticks for the learning curve train sizes
DEFAULT_TRAIN_SIZES = np.linspace(0.1, 1.0, 5)


##########################################################################
# LearningCurve Visualizer
##########################################################################


class LearningCurve(ModelVisualizer):
    """
    Visualizes the learning curve for both test and training data for
    different training set sizes. These curves can act as a proxy to
    demonstrate the implied learning rate with experience (e.g. how much data
    is required to make an adequate model). They also demonstrate if the model
    is more sensitive to error due to bias vs. error due to variance and can
    be used to quickly check if a model is overfitting.

    The visualizer evaluates cross-validated training and test scores for
    different training set sizes. These curves are plotted so that the x-axis
    is the training set size and the y-axis is the score.

    The cross-validation generator splits the whole dataset k times, scores
    are averaged over all k runs for the training subset. The curve plots the
    mean score for the k splits, and the filled in area suggests the
    variability of the cross-validation by plotting one standard deviation
    above and below the mean for each split.

    Parameters
    ----------
    estimator : a scikit-learn estimator
        An object that implements ``fit`` and ``predict``, can be a
        classifier, regressor, or clusterer so long as there is also a valid
        associated scoring metric.

        Note that the object is cloned for each validation.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    groups : array-like, with shape (n_samples,)
        Optional group labels for the samples used while splitting the dataset
        into train/test sets.

    train_sizes : array-like, shape (n_ticks,)
        default: ``np.linspace(0.1,1.0,5)``

        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as
        a fraction of the maximum size of the training set, otherwise it is
        interpreted as absolute sizes of the training sets.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 3-fold cross-validation,
          - integer, to specify the number of folds.
          - An object to be used as a cross-validation generator.
          - An iterable yielding train/test splits.

        see the scikit-learn
        `cross-validation guide <https://bit.ly/2MMQAI7>`_
        for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``. See scikit-learn model evaluation
        documentation for names of possible metrics.

    exploit_incremental_learning : boolean, default: False
        If the estimator supports incremental learning, this will be used to
        speed up fitting for different training set sizes.

    n_jobs : integer, optional
        Number of jobs to run in parallel (default 1).

    pre_dispatch : integer or string, optional
        Number of predispatched jobs for parallel execution (default is
        all). The option can reduce the allocated memory. The string can
        be an expression like '2*n_jobs'.

    shuffle : boolean, optional
        Whether to shuffle training data before taking prefixes of it
        based on``train_sizes``.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`. Used when ``shuffle`` is True.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    train_sizes_ : array, shape = (n_unique_ticks,), dtype int
        Numbers of training examples that has been used to generate the
        learning curve. Note that the number of ticks might be less
        than n_ticks because duplicate entries will be removed.

    train_scores_ : array, shape (n_ticks, n_cv_folds)
        Scores on training sets.

    train_scores_mean_ : array, shape (n_ticks,)
        Mean training data scores for each training split

    train_scores_std_ : array, shape (n_ticks,)
        Standard deviation of training data scores for each training split

    test_scores_ : array, shape (n_ticks, n_cv_folds)
        Scores on test set.

    test_scores_mean_ : array, shape (n_ticks,)
        Mean test data scores for each test split

    test_scores_std_ : array, shape (n_ticks,)
        Standard deviation of test data scores for each test split

    Examples
    --------

    >>> from yellowbrick.model_selection import LearningCurve
    >>> from sklearn.naive_bayes import GaussianNB
    >>> model = LearningCurve(GaussianNB())
    >>> model.fit(X, y)
    >>> model.show()

    Notes
    -----
    This visualizer is essentially a wrapper for the
    ``sklearn.model_selection.learning_curve utility``, discussed in the
    `validation curves <https://bit.ly/2KlumeB>`__
    documentation.

    .. seealso:: The documentation for the
        `learning_curve <https://bit.ly/2Yz9sBB>`__
        function, which this visualizer wraps.
    """

    def __init__(
        self,
        estimator,
        ax=None,
        groups=None,
        train_sizes=DEFAULT_TRAIN_SIZES,
        cv=None,
        scoring=None,
        exploit_incremental_learning=False,
        n_jobs=1,
        pre_dispatch="all",
        shuffle=False,
        random_state=None,
        **kwargs
    ):

        # Initialize the model visualizer
        super(LearningCurve, self).__init__(estimator, ax=ax, **kwargs)

        # Validate the train sizes
        train_sizes = np.asarray(train_sizes)
        if train_sizes.ndim != 1:
            raise YellowbrickValueError(
                "must specify array of train sizes, '{}' is not valid".format(
                    repr(train_sizes)
                )
            )

        # Set the metric parameters to be used later
        self.groups = groups
        self.train_sizes = train_sizes
        self.cv = cv
        self.scoring = scoring
        self.exploit_incremental_learning = exploit_incremental_learning
        self.n_jobs = n_jobs
        self.pre_dispatch = pre_dispatch
        self.shuffle = shuffle
        self.random_state = random_state

    def fit(self, X, y=None):
        """
        Fits the learning curve with the wrapped model to the specified data.
        Draws training and test score curves and saves the scores to the
        estimator.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        Returns
        -------
        self : instance
            Returns the instance of the learning curve visualizer for use in
            pipelines and other sequential transformers.
        """
        # arguments to pass to sk_learning_curve
        sklc_kwargs = {
            key: self.get_params()[key]
            for key in (
                "groups",
                "train_sizes",
                "cv",
                "scoring",
                "exploit_incremental_learning",
                "n_jobs",
                "pre_dispatch",
                "shuffle",
                "random_state",
            )
        }

        # compute the learning curve and store the scores on the estimator
        curve = sk_learning_curve(self.estimator, X, y, **sklc_kwargs)
        self.train_sizes_, self.train_scores_, self.test_scores_ = curve

        # compute the mean and standard deviation of the training data
        self.train_scores_mean_ = np.mean(self.train_scores_, axis=1)
        self.train_scores_std_ = np.std(self.train_scores_, axis=1)

        # compute the mean and standard deviation of the test data
        self.test_scores_mean_ = np.mean(self.test_scores_, axis=1)
        self.test_scores_std_ = np.std(self.test_scores_, axis=1)

        # draw the curves on the current axes
        self.draw()
        return self

    def draw(self, **kwargs):
        """
        Renders the training and test learning curves.
        """
        # Specify the curves to draw and their labels
        labels = ("Training Score", "Cross Validation Score")
        curves = (
            (self.train_scores_mean_, self.train_scores_std_),
            (self.test_scores_mean_, self.test_scores_std_),
        )

        # Get the colors for the train and test curves
        colors = resolve_colors(n_colors=2)

        # Plot the fill betweens first so they are behind the curves.
        for idx, (mean, std) in enumerate(curves):
            # Plot one standard deviation above and below the mean
            self.ax.fill_between(
                self.train_sizes_, mean - std, mean + std, alpha=0.25, color=colors[idx]
            )

        # Plot the mean curves so they are in front of the variance fill
        for idx, (mean, _) in enumerate(curves):
            self.ax.plot(
                self.train_sizes_, mean, "o-", color=colors[idx], label=labels[idx]
            )

        return self.ax

    def finalize(self, **kwargs):
        """
        Add the title, legend, and other visual final touches to the plot.
        """
        # Set the title of the figure
        self.set_title("Learning Curve for {}".format(self.name))

        # Add the legend
        self.ax.legend(frameon=True, loc="best")

        # Set the axis labels
        self.ax.set_xlabel("Training Instances")
        self.ax.set_ylabel("Score")


##########################################################################
# Quick Method
##########################################################################


def learning_curve(
    estimator,
    X,
    y,
    ax=None,
    groups=None,
    train_sizes=DEFAULT_TRAIN_SIZES,
    cv=None,
    scoring=None,
    exploit_incremental_learning=False,
    n_jobs=1,
    pre_dispatch="all",
    shuffle=False,
    random_state=None,
    show=True,
    **kwargs
):
    """
    Displays a learning curve based on number of samples vs training and
    cross validation scores. The learning curve aims to show how a model
    learns and improves with experience.

    This helper function is a quick wrapper to utilize the LearningCurve
    for one-off analysis.

    Parameters
    ----------
    estimator : a scikit-learn estimator
        An object that implements ``fit`` and ``predict``, can be a
        classifier, regressor, or clusterer so long as there is also a valid
        associated scoring metric.

        Note that the object is cloned for each validation.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression;
        None for unsupervised learning.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    groups : array-like, with shape (n_samples,)
        Optional group labels for the samples used while splitting the dataset
        into train/test sets.

    train_sizes : array-like, shape (n_ticks,)
        default: ``np.linspace(0.1,1.0,5)``

        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as
        a fraction of the maximum size of the training set, otherwise it is
        interpreted as absolute sizes of the training sets.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 3-fold cross-validation,
          - integer, to specify the number of folds.
          - An object to be used as a cross-validation generator.
          - An iterable yielding train/test splits.

        see the scikit-learn
        `cross-validation guide <https://bit.ly/2MMQAI7>`_
        for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``. See scikit-learn model evaluation
        documentation for names of possible metrics.

    exploit_incremental_learning : boolean, default: False
        If the estimator supports incremental learning, this will be used to
        speed up fitting for different training set sizes.

    n_jobs : integer, optional
        Number of jobs to run in parallel (default 1).

    pre_dispatch : integer or string, optional
        Number of predispatched jobs for parallel execution (default is
        all). The option can reduce the allocated memory. The string can
        be an expression like '2*n_jobs'.

    shuffle : boolean, optional
        Whether to shuffle training data before taking prefixes of it
        based on``train_sizes``.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`. Used when ``shuffle`` is True.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers. These arguments are
        also passed to the `show()` method, e.g. can pass a path to save the
        figure to.

    Returns
    -------
    visualizer : LearningCurve
        Returns the fitted visualizer.
    """
    # Initialize the visualizer
    oz = LearningCurve(
        estimator,
        ax=ax,
        groups=groups,
        train_sizes=train_sizes,
        cv=cv,
        scoring=scoring,
        n_jobs=n_jobs,
        pre_dispatch=pre_dispatch,
        shuffle=shuffle,
        random_state=random_state,
        exploit_incremental_learning=exploit_incremental_learning,
        **kwargs
    )

    # Fit and show the visualizer
    oz.fit(X, y)
    if show:
        oz.show()
    else:
        oz.finalize()
    return oz



# yellowbrick.contrib.missing.bar
# Missing Values Bar Visualizer
#
# Author:  Nathan Danielsen
# Created: Fri Mar 29 5:17:36 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: bar.py [1443e16] ndanielsen@users.noreply.github.com $

"""
Bar visualizer of missing values by column.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from yellowbrick.style.palettes import color_palette
from .base import MissingDataVisualizer


##########################################################################
## MissingValues Visualizer
##########################################################################


class MissingValuesBar(MissingDataVisualizer):
    """The MissingValues Bar visualizer creates a bar graph that lists the total
    count of missing values for each selected feature column.

    When y targets are supplied to fit, the output is a stacked bar chart where
    each color corresponds to the total NaNs for the feature in that column.

    Parameters
    ----------
    alpha : float, default: 0.5
        A value for bending elments with the background.

    marker : matplotlib marker, default: |
        The marker used for each element coordinate in the plot

    color : string, default: black
        The color for drawing the bar chart when the y targets are not passed to
        fit.

    colors : list, default: None
        The color palette for drawing a stack bar chart when the y targets
        are passed to fit.

    classes : list, default: None
        A list of class names for the legend.
        If classes is None and a y value is passed to fit then the classes
        are selected from the target vector.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : np.array
        The feature labels ranked according to their importance

    classes_ : np.array
        The class labels for each of the target values

    Examples
    --------

    >>> from yellowbrick.contrib.missing import MissingValuesBar
    >>> visualizer = MissingValuesBar()
    >>> visualizer.fit(X, y=y)
    >>> visualizer.show()
    """

    def __init__(self, width=0.5, color=None, colors=None, classes=None, **kwargs):

        if "target_type" not in kwargs:
            kwargs["target_type"] = "single"
        super(MissingValuesBar, self).__init__(**kwargs)
        self.width = width  # the width of the bars
        self.classes_ = classes
        self.ind = None

        # Convert to array if necessary to match estimator.classes_
        if self.classes_ is not None:
            self.classes_ = np.array(classes)

        # Set up classifier score visualization properties
        self.color = color
        if self.classes_ is not None:
            n_colors = len(self.classes_)
        else:
            n_colors = None

        self.colors = color_palette(kwargs.pop("colors", None), n_colors)

    def get_nan_col_counts(self, **kwargs):
        if np.issubdtype(self.X.dtype, np.floating) or np.issubdtype(
                self.X.dtype, np.integer
        ):
            nan_matrix = self.X.astype(np.float64)
        else:
            # where matrix contains strings, handle them
            mask = np.where((self.X == "") | (self.X == 'nan'))
            nan_matrix = np.zeros(self.X.shape)
            nan_matrix[mask] = np.nan

        if self.y is None:
            nan_col_counts = [np.count_nonzero(np.isnan(col)) for col in nan_matrix.T]
            return nan_col_counts

        else:
            # add in counting of np.nan per target y by column
            nan_counts = []
            for target_value in np.unique(self.y):

                indices = np.argwhere(self.y == target_value)
                target_matrix = nan_matrix[indices.flatten()]
                nan_col_counts = np.array(
                    [np.count_nonzero(np.isnan(col)) for col in target_matrix.T]
                )
                nan_counts.append((target_value, nan_col_counts))
            return nan_counts

    def draw(self, X, y, **kwargs):
        """Called from the fit method, this method generated a horizontal bar plot.

        If y is none, then draws a simple horizontal bar chart.
        If y is not none, then draws a stacked horizontal bar chart for each nan count per
        target values.
        """
        nan_col_counts = self.get_nan_col_counts()

        # the x locations for the groups
        self.ind = np.arange(len(self.features_))

        if y is None:
            self.ax.barh(
                self.ind - self.width / 2,
                nan_col_counts,
                self.width,
                color=self.color,
                label=None,
            )
        else:
            self.draw_stacked_bar(nan_col_counts)

    def draw_stacked_bar(self, nan_col_counts):
        """Draws a horizontal stacked bar chart with different colors
        for each count of nan values per label.
        """
        for index, nan_values in enumerate(nan_col_counts):
            label, nan_col_counts = nan_values

            if index == 0:
                # first draw should be at zero
                bottom_chart = np.zeros(nan_col_counts.shape)

            # if features passed in then, label as such
            if self.classes_ is not None:
                label = self.classes_[index]

            color = self.colors[index]

            self.ax.barh(
                self.ind - self.width / 2,
                nan_col_counts,
                self.width,
                color=color,
                label=label,
                left=bottom_chart,
            )

            # keep track of counts to build on stacked
            bottom_chart = nan_col_counts

    def finalize(self, **kwargs):
        """
        Sets a title and x-axis labels and adds a legend. Also ensures that the
        y tick values are correctly set to feature names.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title
        self.set_title("Count of Missing Values by Column")
        tick_locations = np.arange(
            len(self.features_)
        )  # the x locations for the groups
        self.ax.set_yticks(tick_locations)
        self.ax.set_yticklabels(self.get_feature_names())
        # Remove the ticks from the graph
        self.ax.set_xlabel("Count")

        self.ax.legend(loc="best")


##########################################################################
## Quick Method
##########################################################################


def missing_bar(X, y=None, ax=None, classes=None, width=0.5, color="black", **kwargs):
    """The MissingValues Bar visualizer creates a bar graph that lists the total
    count of missing values for each selected feature column.

    When y targets are supplied to fit, the output is a stacked bar chart where
    each color corresponds to the total NaNs for the feature in that column.

    Parameters
    ----------
    alpha : float, default: 0.5
        A value for bending elments with the background.

    marker : matplotlib marker, default: |
        The marker used for each element coordinate in the plot

    color : string, default: black
        The color for drawing the bar chart when the y targets are not passed to
        fit.

    colors : list, default: None
        The color pallette for drawing a stack bar chart when the y targets
        are passed to fit.

    classes : list, default: None
        A list of class names for the legend.
        If classes is None and a y value is passed to fit then the classes
        are selected from the target vector.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : np.array
        The feature labels ranked according to their importance

    classes_ : np.array
        The class labels for each of the target values

    Examples
    --------

    >>> from yellowbrick.contrib.missing import missing_bar
    >>> visualizer = missing_bar(X, y=y)
    """
    # Instantiate the visualizer
    visualizer = MissingValuesBar(
        ax=ax, classes=classes, width=width, color=color, **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y)
    visualizer.show()

    # Return the axes object on the visualizer
    return visualizer.ax



# yellowbrick.regressor
# Visualizers for Regression analysis and diagnostics
#
# Author:   Benjamin Bengfort
# Created:  Mon Mar 06 12:23:55 2017 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [7d3f5e6] benjamin@bengfort.com $

"""
Visualizers for Regression analysis and diagnostics, particularly
visualizations related to evaluating Scikit-Learn regressor models.
"""

##########################################################################
## Imports
##########################################################################

## Hoist visualizers into the regressor namespace
from .base import *
from .residuals import *
from .prediction_error import *
from .alphas import *
from .influence import *



# yellowbrick.regressor.alphas
# Implements alpha selection visualizers for regularization
#
# Author:   Benjamin Bengfort
# Author:   Rebecca Bilbro
# Created:  Mon Mar 06 19:22:07 2017 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: alphas.py [7d3f5e6] benjamin@bengfort.com $

"""
Implements alpha selection visualizers for regularization
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from functools import partial

from yellowbrick.exceptions import YellowbrickTypeError
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.regressor.base import RegressionScoreVisualizer

from sklearn.model_selection import cross_val_score

## Packages for export
__all__ = ["AlphaSelection", "ManualAlphaSelection"]


##########################################################################
## AlphaSelection Visualizer
##########################################################################


class AlphaSelection(RegressionScoreVisualizer):
    """
    The Alpha Selection Visualizer demonstrates how different values of alpha
    influence model selection during the regularization of linear models.
    Generally speaking, alpha increases the affect of regularization, e.g. if
    alpha is zero there is no regularization and the higher the alpha, the
    more the regularization parameter influences the final model.

    Regularization is designed to penalize model complexity, therefore the
    higher the alpha, the less complex the model, decreasing the error due to
    variance (overfit). Alphas that are too high on the other hand increase
    the error due to bias (underfit). It is important, therefore to choose an
    optimal Alpha such that the error is minimized in both directions.

    To do this, typically you would you use one of the "RegressionCV" models
    in Scikit-Learn. E.g. instead of using the ``Ridge`` (L2) regularizer, you
    can use ``RidgeCV`` and pass a list of alphas, which will be selected
    based on the cross-validation score of each alpha. This visualizer wraps
    a "RegressionCV" model and visualizes the alpha/error curve. Use this
    visualization to detect if the model is responding to regularization, e.g.
    as you increase or decrease alpha, the model responds and error is
    decreased. If the visualization shows a jagged or random plot, then
    potentially the model is not sensitive to that type of regularization and
    another is required (e.g. L1 or ``Lasso`` regularization).

    Parameters
    ----------

    estimator : a Scikit-Learn regressor
        Should be an instance of a regressor, and specifically one whose name
        ends with "CV" otherwise a will raise a YellowbrickTypeError exception
        on instantiation. To use non-CV regressors see: ``ManualAlphaSelection``.
        If the estimator is not fitted, it is fit when the visualizer is fitted,
        unless otherwise specified by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    is_fitted : bool or str, default='auto'
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If 'auto' (default), a helper method will check if the estimator
        is fitted before fitting it again.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Examples
    --------

    >>> from yellowbrick.regressor import AlphaSelection
    >>> from sklearn.linear_model import LassoCV
    >>> model = AlphaSelection(LassoCV())
    >>> model.fit(X, y)
    >>> model.show()

    Notes
    -----

    This class expects an estimator whose name ends with "CV". If you wish to
    use some other estimator, please see the ``ManualAlphaSelection``
    Visualizer for manually iterating through all alphas and selecting the
    best one.

    This Visualizer hooks into the Scikit-Learn API during ``fit()``. In
    order to pass a fitted model to the Visualizer, call the ``draw()`` method
    directly after instantiating the visualizer with the fitted model.

    Note, each "RegressorCV" module has many different methods for storing
    alphas and error. This visualizer attempts to get them all and is known
    to work for RidgeCV, LassoCV, LassoLarsCV, and ElasticNetCV. If your
    favorite regularization method doesn't work, please submit a bug report.

    For RidgeCV, make sure ``store_cv_values=True``.
    """

    def __init__(self, estimator, ax=None, is_fitted="auto", **kwargs):

        # Check to make sure this is a "RegressorCV"
        name = estimator.__class__.__name__
        if not name.endswith("CV"):
            raise YellowbrickTypeError(
                (
                    "'{}' is not a CV regularization model;"
                    " try ManualAlphaSelection instead."
                ).format(name)
            )

        # Set the store_cv_values parameter on RidgeCV
        if "store_cv_values" in estimator.get_params().keys():
            estimator.set_params(store_cv_values=True)

        # Call super to initialize the class
        super(AlphaSelection, self).__init__(estimator, ax=ax, **kwargs)

    def fit(self, X, y, **kwargs):
        """
        A simple pass-through method; calls fit on the estimator and then
        draws the alpha-error plot.
        """
        # Fit the underlying model
        super(AlphaSelection, self).fit(X, y, **kwargs)

        # Draw the alpha to error curve
        self.draw()
        return self

    def draw(self):
        """
        Draws the alpha plot based on the values on the estimator.
        """
        # Search for the correct parameters on the estimator.
        alphas = self._find_alphas_param()
        errors = self._find_errors_param()

        alpha = self.estimator.alpha_  # Get decision from the estimator
        name = self.name[:-2].lower()  # Remove the CV from the label

        # Plot the alpha against the error
        self.ax.plot(alphas, errors, label=name)

        # Draw a dashed vline at the alpha
        label = "$\\alpha={:0.3f}$".format(alpha)
        self.ax.axvline(alpha, color="k", linestyle="dashed", label=label)

        return self.ax

    def finalize(self):
        """
        Prepare the figure for rendering by setting the title as well as the
        X and Y axis labels and adding the legend.
        """
        # Set the title
        self.set_title("{} Alpha Error".format(self.name))

        # Set the x and y labels
        self.ax.set_xlabel("alpha")
        self.ax.set_ylabel("error (or score)")

        # Set the legend
        self.ax.legend(loc="best", frameon=True)

    def _find_alphas_param(self):
        """
        Searches for the parameter on the estimator that contains the array of
        alphas that was used to produce the error selection. If it cannot find
        the parameter then a YellowbrickValueError is raised.
        """

        # NOTE: The order of the search is very important!
        for attr in ("cv_alphas_", "alphas_", "alphas"):
            try:
                return getattr(self.estimator, attr)
            except AttributeError:
                continue

        raise YellowbrickValueError(
            "could not find alphas param on {} estimator".format(
                self.estimator.__class__.__name__
            )
        )

    def _find_errors_param(self):
        """
        Searches for the parameter on the estimator that contains the array of
        errors that was used to determine the optimal alpha. If it cannot find
        the parameter then a YellowbrickValueError is raised.
        """

        # NOTE: The order of the search is very important!
        if hasattr(self.estimator, "mse_path_"):
            return self.estimator.mse_path_.mean(1)

        if hasattr(self.estimator, "cv_values_"):
            return self.estimator.cv_values_.mean(0)

        raise YellowbrickValueError(
            "could not find errors param on {} estimator".format(
                self.estimator.__class__.__name__
            )
        )


##########################################################################
## ManualAlphaSelection Visualizer
##########################################################################


class ManualAlphaSelection(AlphaSelection):
    """
    The ``AlphaSelection`` visualizer requires a "RegressorCV", that is a
    specialized class that performs cross-validated alpha-selection on behalf
    of the model. If the regressor you wish to use doesn't have an associated
    "CV" estimator, or for some reason you would like to specify more control
    over the alpha selection process, then you can use this manual alpha
    selection visualizer, which is essentially a wrapper for
    ``cross_val_score``, fitting a model for each alpha specified.

    Parameters
    ----------

    estimator : an unfitted Scikit-Learn regressor
        Should be an instance of an unfitted regressor, and specifically one
        whose name doesn't end with "CV". The regressor must support a call to
        ``set_params(alpha=alpha)`` and be fit multiple times. If the
        regressor name ends with "CV" a ``YellowbrickValueError`` is raised.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    alphas : ndarray or Series, default: np.logspace(-10, 2, 200)
        An array of alphas to fit each model with

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - An object to be used as a cross-validation generator.
        - An iterable yielding train, test splits.

        This argument is passed to the
        ``sklearn.model_selection.cross_val_score`` method to produce the
        cross validated score for each alpha.

    scoring : string, callable or None, optional, default: None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

        This argument is passed to the
        ``sklearn.model_selection.cross_val_score`` method to produce the
        cross validated score for each alpha.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Examples
    --------

    >>> from yellowbrick.regressor import ManualAlphaSelection
    >>> from sklearn.linear_model import Ridge
    >>> model = ManualAlphaSelection(
    ...     Ridge(), cv=12, scoring='neg_mean_squared_error'
    ... )
    ...
    >>> model.fit(X, y)
    >>> model.show()

    Notes
    -----

    This class does not take advantage of estimator-specific searching and is
    therefore less optimal and more time consuming than the regular
    "RegressorCV" estimators.
    """

    def __init__(
        self, estimator, ax=None, alphas=None, cv=None, scoring=None, **kwargs
    ):

        # Check to make sure this is not a "RegressorCV"
        name = estimator.__class__.__name__
        if name.endswith("CV"):
            raise YellowbrickTypeError(
                (
                    "'{}' is a CV regularization model;" " try AlphaSelection instead."
                ).format(name)
            )

        # Call super to initialize the class
        super(AlphaSelection, self).__init__(estimator, ax=ax, **kwargs)

        # Set manual alpha selection parameters
        if alphas is not None:
            self.alphas = alphas
        else:
            self.alphas = np.logspace(-10, -2, 200)
        self.errors = None
        self.score_method = partial(cross_val_score, cv=cv, scoring=scoring)

    def fit(self, X, y, **args):
        """
        The fit method is the primary entry point for the manual alpha
        selection visualizer. It sets the alpha param for each alpha in the
        alphas list on the wrapped estimator, then scores the model using the
        passed in X and y data set. Those scores are then aggregated and
        drawn using matplotlib.
        """
        self.errors = []
        for alpha in self.alphas:
            self.estimator.set_params(alpha=alpha)
            scores = self.score_method(self.estimator, X, y)
            self.errors.append(scores.mean())

        # Convert errors to an ND array and draw
        self.errors = np.array(self.errors)
        self.draw()

        # Always make sure to return self from fit
        return self

    def draw(self):
        """
        Draws the alphas values against their associated error in a similar
        fashion to the AlphaSelection visualizer.
        """
        # Plot the alpha against the error
        self.ax.plot(self.alphas, self.errors, label=self.name.lower())

        # Draw a dashed vline at the alpha with maximal error
        alpha = self.alphas[np.where(self.errors == self.errors.max())][0]
        label = "$\\alpha_{{max}}={:0.3f}$".format(alpha)
        self.ax.axvline(alpha, color="k", linestyle="dashed", label=label)

        # Draw a dashed vline at the alpha with minimal error
        alpha = self.alphas[np.where(self.errors == self.errors.min())][0]
        label = "$\\alpha_{{min}}={:0.3f}$".format(alpha)
        self.ax.axvline(alpha, color="k", linestyle="dashed", label=label)

        return self.ax


##########################################################################
## Quick Methods
##########################################################################


def alphas(estimator, X, y=None, ax=None, is_fitted="auto", show=True, **kwargs):
    """Quick Method:
    The Alpha Selection Visualizer demonstrates how different values of alpha
    influence model selection during the regularization of linear models.
    Generally speaking, alpha increases the affect of regularization, e.g. if
    alpha is zero there is no regularization and the higher the alpha, the
    more the regularization parameter influences the final model.

    Parameters
    ----------

    estimator : a Scikit-Learn regressor
        Should be an instance of a regressor, and specifically one whose name
        ends with "CV" otherwise a will raise a YellowbrickTypeError exception
        on instantiation. To use non-CV regressors see: ``ManualAlphaSelection``.
        If the estimator is not fitted, it is fit when the visualizer is fitted,
        unless otherwise specified by ``is_fitted``.

    X  : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features.

    y  : ndarray or Series of length n
        An array or series of target values.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    is_fitted : bool or str, default='auto'
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If 'auto' (default), a helper method will check if the estimator
        is fitted before fitting it again.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    visualizer : AlphaSelection
        Returns the alpha selection visualizer
    """
    # Instantiate the visualizer
    visualizer = AlphaSelection(estimator, ax, is_fitted=is_fitted, **kwargs)

    visualizer.fit(X, y)
    visualizer.score(X, y)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer


def manual_alphas(
    estimator,
    X,
    y=None,
    ax=None,
    alphas=None,
    cv=None,
    scoring=None,
    show=True,
    **kwargs
):
    """Quick Method:
    The Manual Alpha Selection Visualizer demonstrates how different values of alpha
    influence model selection during the regularization of linear models.
    Generally speaking, alpha increases the affect of regularization, e.g. if
    alpha is zero there is no regularization and the higher the alpha, the
    more the regularization parameter influences the final model.

    Parameters
    ----------

    estimator : an unfitted Scikit-Learn regressor
        Should be an instance of an unfitted regressor, and specifically one
        whose name doesn't end with "CV". The regressor must support a call to
        ``set_params(alpha=alpha)`` and be fit multiple times. If the
        regressor name ends with "CV" a ``YellowbrickValueError`` is raised.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    alphas : ndarray or Series, default: np.logspace(-10, 2, 200)
        An array of alphas to fit each model with

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - An object to be used as a cross-validation generator.
        - An iterable yielding train, test splits.

        This argument is passed to the
        ``sklearn.model_selection.cross_val_score`` method to produce the
        cross validated score for each alpha.

    scoring : string, callable or None, optional, default: None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

        This argument is passed to the
        ``sklearn.model_selection.cross_val_score`` method to produce the
        cross validated score for each alpha.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    visualizer : AlphaSelection
        Returns the alpha selection visualizer
    """
    # Instantiate the visualizer
    visualizer = ManualAlphaSelection(
        estimator, ax, alphas=alphas, scoring=scoring, cv=cv, **kwargs
    )

    visualizer.fit(X, y)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer



# yellowbrick.utils.target
# Helper functions related to the target variable.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Thu Dec 27 20:16:18 2018 -0500
#
# Copyright (C) 2018 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: target.py [899c88a] benjamin@bengfort.com $

"""
Helper functions related to the target variable.
"""

##########################################################################
## Imports and Module Variables
##########################################################################

import numpy as np

from enum import Enum
from sklearn.utils.multiclass import type_of_target
from yellowbrick.exceptions import YellowbrickValueError


__all__ = ["MAX_DISCRETE_CLASSES", "TargetType", "target_color_type"]

MAX_DISCRETE_CLASSES = 12


class TargetType(Enum):
    """Constants for defining target colors by input type"""

    AUTO = "auto"
    SINGLE = "single"
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"
    UNKNOWN = "unknown"

    @classmethod
    def validate(klass, val):
        if isinstance(val, klass):
            return

        try:
            klass(val)
        except ValueError:
            raise YellowbrickValueError("unknown target color type '{}'".format(val))

    def __eq__(self, other):
        if isinstance(other, str):
            try:
                return TargetType(other.lower()) == self
            except ValueError:
                return False
        return super(TargetType, self).__eq__(other)


##########################################################################
## Helper Functions
##########################################################################


def target_color_type(y):
    """
    Determines the type of color space that will best represent the target
    variable y, e.g. either a discrete (categorical) color space or a
    continuous color space that requires a colormap. This function can handle
    both 1D or column vectors as well as multi-output targets.

    Parameters
    ----------
    y : array-like
        Must be a valid array-like data structure that can be passed to a
        scikit-learn supervised estimator.

    Returns
    -------
    color_type : string
        One of:

        * 'discrete': `y` is either a binary target or a multiclass target
          with <= 12 discrete classes.
        * 'continuous': `y` is an array-like of floats that are not all
          integers or a multiclass target with > 12 discrete classes.
        * 'unknown': `y` is array-like but none of the above. For example
          a multilabel-indicator or a 3D array. No exception is raised.
    """
    if y is None or len(np.unique(y)) == 1:
        return TargetType.SINGLE

    ttype = type_of_target(y)

    if ttype.startswith("continuous"):
        return TargetType.CONTINUOUS

    if ttype.startswith("binary"):
        return TargetType.DISCRETE

    if ttype.startswith("multiclass"):
        if len(np.unique(y)) > MAX_DISCRETE_CLASSES:
            return TargetType.CONTINUOUS
        return TargetType.DISCRETE

    return TargetType.UNKNOWN
